/*!
 * @file fas_filesystem.h
 * @brief This is the header file for the Fastec Filesystem library.
 *
 * Copyright 2013-2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */
#ifndef FAS_FILESYSTEM_H_
#define FAS_FILESYSTEM_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>

#include <memory>
#include <mutex>
#include <string>
#include <vector>

namespace FAS_FILESYSTEM
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! This macro is how we create condensed signature information.
#define MAKE_SIGNATURE( a, b, c, d ) \
    ((((uint32_t)(d))<<24) | (((uint32_t)(c))<<16) | \
     (((uint32_t)(b))<<8)  | ((uint32_t)(a)))

//! @name Signatures
//! These are the signatures used for the file system headers.
//! @{
const uint32_t DRIVE_SIG        = MAKE_SIGNATURE('L', 'R', 'D', 'V');   //!< A Fastec Filesystem.

const uint32_t AVAIL_PART_SIG   = MAKE_SIGNATURE('A', 'V', 'A', 'L');   //!< An available partition.

const uint32_t LR_PART_SIG      = MAKE_SIGNATURE('P', 'A', 'R', 'T');   //!< A "standard" long recording partition.
const uint32_t LR_CAPTURE_SIG   = MAKE_SIGNATURE('C', 'A', 'P', 'T');   //!< A "standard" long recording capture.

const uint32_t STOR_PART_SIG    = MAKE_SIGNATURE('S', 'T', 'O', 'R');   //!< A storage partition.
const uint32_t STOR_CAPTURE_SIG = MAKE_SIGNATURE('C', 'S', 'T', 'R');   //!< A storage capture (CAP file).

const uint32_t CT_GT_HDR_ID     = MAKE_SIGNATURE('C', 'T', 'G', 'T');   //!< A correction/gain table header.

//
//! @}

//! The various types of bookmarks we support.
typedef enum
{
    FIBK_EMPTY  = MAKE_SIGNATURE('M', 'P', 'T', 'Y'),       //!< This bookmark entry is empty.
    FIBK_USER   = MAKE_SIGNATURE('U', 'S', 'E', 'R'),       //!< This indicates a USER bookmark.
    FIBK_DSPD   = MAKE_SIGNATURE('D', 'S', 'P', 'D')        //!< This is a "dual-speed" bookmark entry.
} FI_BOOKMARK_TYPE;

//! The maximum allowed size for the bookmark name field, including the null terminator.
const int FIBK_NAME_LEN         = 16;

//! The maximum allowed size (in bytes) for the capture bookmark table.
const uint32_t CAPTURE_BOOKMARK_SIZE = 4096;

//! The maximum allowed size (in bytes) for the capture metadata.
const uint32_t CAPTURE_METADATA_SIZE  = 4096;

//! 512 bytes per sector.
const int SECTOR_LEN            = 512;

//! How many sectors it takes to hold these bytes.
#define BYTES_TO_SECTORS( b )   (((b) + FAS_FILESYSTEM::SECTOR_LEN - 1) / FAS_FILESYSTEM::SECTOR_LEN)

//! 8 sectors per page.
const int SECTORS_PER_PAGE      = 8;

//! 4096 bytes per page.
const int SSD_PAGE_LEN          = ((SECTORS_PER_PAGE) * (SECTOR_LEN));

//! @name Frame metadata word offsets.
//! @{
const int FMETA_TIME_SECONDS    = 0;        //!< The seconds component of the frame time.
const int FMETA_TIME_USECS      = 1;        //!< The microseconds component of the frame time.
const int FMETA_MARKERS         = 2;        //!< Location of the marker information.
const int FMETA_HW_NUM          = 3;        //!< The hardware frame number.
const int FMETA_FIFO_LEVELS     = 4;        //!< The FIFO levels at the time of the recording.
const int FMETA_RECORDING_ID    = 5;        //!< The software supplied recording ID.
const int FMETA_UNUSED          = 6;        //!< Currently unused, set to zero.
const int FMETA_VTRIG_INFO      = 7;        //!< Video trigger information.
//
//! @}

//! Mask to extract microseconds from frame metadata word that contains some other flags.
const uint32_t FMETA_TIME_USEC_MASK = 0xfffff;

//! States of a CAPTURE header
enum CAPTUREHD_STATE
{
    CAPTUREHD_ALLOCATED,                    //!< The capture has been allocated, but not initialized.
    CAPTUREHD_READY,                        //!< The capture is ready to be written to.
    CAPTUREHD_CAPTURING,                    //!< The capture is currently being written to.
    CAPTUREHD_COMPLETE,                     //!< The capture has been written.
    CAPTUREHD_APPENDING                     //!< The capture is in append mode.
};

//! @name Flags in the CAPTURE header
//! @{
const int CAPHD_FLAG_RECONSTRUCTED  = 0x01; //!< Reconstructed start+end after a crash.
const int CAPHD_FLAG_APPENDMODE     = 0x02; //!< ROC or BROC mode recording.
//
//! @}

//! How many drive partitions we can store on a drive.
const int MAX_DPARTS        = 40;

//! How long can the name of the drive be, not including the null termination.
const int MAX_DRV_NAME_LEN  = 63;

//! How long can the name of the storage capture be, not including the null termination.
const int MAX_CAP_NAME_LEN  = 63;

//! How big is the ver1 DRIVE_HD structure. Essentially this
//! is: sizeof(DRIVE_HD) - (MAX_DRV_NAME_LEN + 1).
const int DRIVE_HDR_VER1_LEN = (8 + 2 + (2 * 3) + (8 * 41));

//! Minimum size of a storage capture (CAP), in sectors.
const uint32_t STOR_CAPTURE_MIN_SECTORS = 524288;

#pragma pack(1)
/*!
 * @brief The signature structure on all Fastec Filesystem headers.
 */
typedef struct FI_SIG
{
    uint32_t    signature;                  //!< Signature for this type of header.
    uint16_t    size;                       //!< Size in bytes of data beyond the header.
    uint16_t    checksum;                   //!< Checksum covering size bytes.
} FI_SIG;

/*!
 * @brief Fastec Filesystem drive header.
 *
 * This drive header uniquely identifies the drive as having
 * the Fastec Filesystem format.
 */
typedef struct DRIVE_HD
{
    uint64_t    drive_size;                 // The drive size on which this format was based.

    // The size of allocation units in sectors.
    // This is for info and future use, since we code with the defined
    // constant SECTORS_PER_PAGE.
    uint16_t    alloc_unit;

    uint16_t    placeholder[3];             // Reserved for future use, set to zero.
    uint64_t    dpart_lba[MAX_DPARTS + 1];  // Array of drive partition LBAs, ending with zero.
    uint8_t     name[MAX_DRV_NAME_LEN + 1]; // User provided name for the drive, null terminated.
} DRIVE_HD;

//! The FIBK_USER entry.
typedef struct FIBK_USER_ENTRY
{
    uint32_t    frame;                  // The frame of interest in this capture, relative to "first_frame_num".
} FIBK_USER_ENTRY;

//! The FIBK_DSPD entry.
typedef struct FIBK_DSPD_ENTRY
{
    uint32_t    lr_first_frame;         // The first frame in the LR capture, relative to "first_frame_num".
    uint32_t    count;                  // How many frames, including the first LR frame, are in the range.
    uint64_t    hs_cap_lba;             // Location of the HS capture header.
} FIBK_DSPD_ENTRY;

/*!
 * @brief The generic bookmark table entry.
 */
typedef struct FI_BOOKMARK
{
    uint32_t type;                      // Type of bookmark, one of FI_BOOKMARK_TYPE.
    char     name[FIBK_NAME_LEN];       // Name given to the bookmark.
    union
    {
        FIBK_USER_ENTRY user;
        FIBK_DSPD_ENTRY dspd;
    } entry;
} FI_BOOKMARK;

//! The maximum number of bookmark entries that will fit within CAPTURE_BOOKMARK_SIZE.
const int MAX_BOOKMARK_ENTRIES = (CAPTURE_BOOKMARK_SIZE / sizeof(FI_BOOKMARK));

/*!
 * @brief The correction/gain table header.
 */
typedef struct CT_GT_HDR
{
    uint32_t magic;                     // Our magic number.
    uint32_t ct_bytes;                  // Size of Correction Table area.
    uint32_t ct_offset;                 // Byte offset from start of this structure's sector to CT area, sector aligned.
    uint32_t gt_bytes;                  // Size of Gain Table area.
    uint32_t gt_offset;                 // Byte offset from start of this structure's sector to GT area, sector aligned.
} CT_GT_HDR;

/*!
 * @brief The generic portion of a capture header. All capture
 *        headers will have this at the beginning.
 */
typedef struct FI_CAPTURE_HDR
{
    // We need the linkage to the parent and siblings to traverse the file system.
    uint64_t dpart_lba;                 // Header location of drive partition to which this capture belongs.
    uint64_t previous_lba;              // Location of previous capture header.
    uint64_t next_lba;                  // Location of next capture header.
} FI_CAPTURE_HDR;

//! @name Partition Header Definitions
//! @{

/*!
 * @brief "Standard" long record drive partition header for holding captures.
 *
 * If we have different types of partitions, e.g. one to hold CAP files,
 * they can have a different signature. A partition can have multiple
 * child captures linked together, which all share the common partition
 * image and frame metadata regions. The minimum frame size we used to
 * decide how much frame metadata to allocate is probably unnecessary
 * since all we really care about is whether there is space left to
 * support the next capture, but it may be useful while getting this running.
 */
typedef struct LR_PARTITION_HDR
{
    uint64_t fmeta_lba;                 // Preallocated frame metadata region shared by all the captures within this partition. 0 if unused.
    uint64_t fmeta_sectors;             // Size of frame metadata region. 0 if unused.
    uint64_t image_lba;                 // Location of the start of image data corresponding to fmeta_lba.
    uint64_t image_sectors;             // Image area size on which the fmeta allocation was based. Appended captures will use this to determine remaining usable space.
    uint64_t capturehd_lba;             // Location of the first capture header for this partition. Capture metadata must fit between capturehd_lba and image_lba.
    uint32_t min_frame_size;            // Minimum frame size on which the fmeta allocation was based. If not permitting append, this can be the actual frame size.

    uint8_t  reserved_sectors;          // Always 1 for now.
    uint8_t  placeholder[3];            // Reserved for future use.
} LR_PARTITION_HDR;

/*!
 * @brief "Storage" partition header.
 */
typedef struct STOR_PARTITION_HDR
{
    uint64_t meta_lba;                  // Preallocated metadata region shared by all the captures within this partition. 0 if unused.
    uint64_t meta_sectors;              // Size of metadata region. 0 if unused.
    uint64_t start_lba;                 // Location of the start of the data space for the partition.
    uint64_t data_sectors;              // Number of sectors in the data space for this partition.
    uint64_t capturehd_lba;             // Location of the first capture header for this partition. This will be somewhere between start_lba and (start_lba+data_sectors).

    uint8_t  reserved_sectors;          // Always 1 for now.
    uint8_t  placeholder[3];            // Reserved for future use.
} STOR_PARTITION_HDR;

/*!
 * @brief "Available" partition header.
 */
typedef struct AVAIL_PARTITION_HDR
{
    uint64_t start_lba;                 // Location of the start of the data space for the partition.
    uint64_t data_sectors;              // Number of sectors in the data space for this partition.

    uint8_t  reserved_sectors;          // Always 1 for now.
    uint8_t  placeholder[3];            // Reserved for future use.
} AVAIL_PARTITION_HDR;

//
//! @}

//! @name Capture Header Definitions.
//! @{

/*!
 * @brief Describes a single "standard" long record capture.
 *
 * This will describe a basic, BROC, or ROC recording. Some of the metadata from
 * the recording is duplicated in the header to allow for easier display to the user.
 */
typedef struct LR_CAPTURE_HDR
{
    FI_CAPTURE_HDR  link;

    // This describes where the image data and frame metadata are on the drive.
    // They are set if the header is in the ALLOCATED state or later.
    uint64_t fmeta_lba;                 // Start of per-frame metadata for this capture. 0 if unused.
    uint64_t fmeta_sectors;             // Number of sectors allocated for frame metadata for this capture. 0 if unused.
    uint64_t image_lba;                 // Start of image data for this capture.
    uint64_t image_sectors;             // Number of sectors allocated for image data.

    // These are other metadata structures for the capture.
    // These are set if the capture header is in the READY state or later.
    uint32_t bk_bytes;                  // Size of bookmark area.
    uint32_t bk_offset;                 // Byte offset from start of capture header sector to bookmark area.
    uint32_t pm_bytes;                  // Size of this capture's memory partition metadata.
    uint32_t pm_offset;                 // Byte offset from start of capture header sector to PM area.
    uint32_t ct_bytes;                  // Size of Correction/Gain Table area.
    uint32_t ct_offset;                 // Byte offset from start of capture header sector to CT/GT area.
    uint32_t reserved_sectors;          // Total capture reserved sectors including this header.

    // Some of this is redundant with the memory partition metadata. Having it here lets us
    // write the completion information as a single sector once the capture is done.
    // We update it at the end of a capture then when memory partition metadata is loaded, we
    // will update the pm_metadata in memory based on what is here.
    // We also need some of this information to determine what space can be freed up if we want
    // to break a capture region into a used area and a free area after a capture stops.

    // These are set for a header that is in the CAPTURING state or later and for the
    // APPENDING state or later.
    uint32_t recording_id;              // ID for this recording.
    uint32_t frame_size;                // Frame size in bytes.
    uint32_t frame_sectors;             // Frame size in sectors.
    uint32_t frame_count;               // Number of frames allocated => floor((image_sectors * 512) / frame_size).

    // These are set for a header that is in the COMPLETED state.
    uint32_t num_frames;                // Number of frames actually captured, once capture is done.
    uint32_t first_frame_num;           // Oldest frame, number relative to first frame of the capture area.
    uint32_t trigger_frame_num;         // Trigger frame, number is relative to first_frame_num.
    uint32_t trigger_time_s;            // Trigger time seconds.
    uint32_t trigger_time_ns;           // Trigger time ns past the second (but only usec precision).

    // State of this capture: ALLOCATED, READY, CAPTURING, COMPLETED
    // ALLOCATED - Frame metadata and image allocations are defined
    // READY     - Params describe a capture but recording is not
    //             started.  Shouldn't have any need for this on the
    //             disk, but useful for dragging parameters around.
    // CAPTURING - Saved when starting a normal capture. See above.
    // APPENDING - Saved when starting a ROC/BROC capture. See above.
    // COMPLETE  - Saved when capture is complete or after post-crash
    //             recovery.
    uint8_t state;

    uint8_t flags;                      // Set a flag if we reconstructed the recording from the CAPTURING state.

    uint8_t placeholder[2];             // Keep the size a multiple of 32-bit words.

} LR_CAPTURE_HDR;

/*!
 * @brief Describes a single "storage" capture.
 *
 * A storage capture can hold one Fastec "CAP" file. We duplicate some of the metadata
 * from the CAP information in the header to make displaying to the user easier.
 */
typedef struct STOR_CAPTURE_HDR
{
    FI_CAPTURE_HDR  link;

    uint64_t data_lba;                  // Start of data space for this capture.
    uint64_t data_sectors;              // Number of sectors allocated for data.
    uint32_t reserved_sectors;          // Total capture reserved sectors including this header.

    // Some metadata associated with this capture.
    uint64_t used_sectors;              // Number of sectors used for the data.
    uint32_t trigger_time_s;            // Trigger time seconds.
    uint32_t trigger_time_ns;           // Trigger time ns past the second (but only usec precision).
    char name[MAX_CAP_NAME_LEN + 1];    // Name associated with the capture, null terminated.

    // State of this capture: ALLOCATED, CAPTURING, COMPLETED
    // ALLOCATED - Space has been allocated, but contains no data.
    // CAPTURING - Space is being written to.
    // COMPLETE  - Saved when capture is complete.
    uint8_t state;

    uint8_t flags;                      // Set a flag if we reconstructed the recording from the CAPTURING state.

    uint8_t placeholder[2];             // Keep the size a multiple of 32-bit words.

} STOR_CAPTURE_HDR;

//
//! @}

#pragma pack(0)

//! A collection of all possible capture headers.
typedef union CAPTURE_UNION
{
    FI_CAPTURE_HDR      link;           //!< Part of all capture headers.
    LR_CAPTURE_HDR      lr_cap;         //!< Information for a Long Record Capture.
    STOR_CAPTURE_HDR    stor_cap;       //!< Information for a Storage Capture.
} CAPTURE_UNION;

//! A collection of all possible partition headers.
typedef union PARTITION_UNION
{
    AVAIL_PARTITION_HDR avail_part;     //!< Information for an Available Partition.
    LR_PARTITION_HDR    lr_part;        //!< Information for a Long Record Partition.
    STOR_PARTITION_HDR  stor_part;      //!< Information for a Storage Partition.
} PARTITION_UNION;

//! Specific details about a particular capture.
typedef struct CAPTURE_DATA
{
    uint32_t type;                          //!< The type will tell us what the hdr union will contain.
    uint64_t lba;                           //!< LBA for this header.
    CAPTURE_UNION hdr;                      //!< Collection of capture headers, identified by "type".
} CAPTURE_DATA;
typedef std::unique_ptr<CAPTURE_DATA> CAP_UNION_PTR;            //!< Utility typedef to make programming a little easier.
typedef std::vector<CAPTURE_DATA>     CAP_UNION_VEC;            //!< Utility typedef to make programming a little easier.

//! Specific details about a particular partition.
typedef struct PARTITION_DATA
{
    uint32_t type;                          //!< The type will tell us what the hdr union will contain.
    uint64_t lba;                           //!< LBA for this header.
    PARTITION_UNION hdr;                    //!< Collection of partition headers, identified by "type".
} PARTITION_DATA;
typedef std::unique_ptr<PARTITION_DATA> PARTITION_DATA_PTR;     //!< Utility typedef to make programming a little easier.

//! Information about a particular partition.
typedef struct PARTITION_INFO
{
    PARTITION_DATA  part;                   //!< The partition specific information.
    CAP_UNION_VEC  caps;                    //!< The vector containing the information of one or more captures.
} PARTITION_INFO;
typedef std::unique_ptr<PARTITION_INFO> PARTITION_INFO_PTR;     //!< Utility typedef to make programming a little easier.
typedef std::vector<PARTITION_INFO>     PARTITION_INFO_VEC;     //!< Utility typedef to make programming a little easier.

//! Information used in the creation of a capture header.
typedef struct LAYOUT_CAPTURE
{
    //! What type of capture are we laying down.
    uint32_t signature;

    //! How many sectors to pre-assign to each capture. 0 means the size will be filled
    //! in at capture time.
    uint64_t cap_size;

    //! How many bytes to reserve for the correction/gain table(s) information.
    uint32_t max_ct_size;
} LAYOUT_CAPTURE;
typedef std::vector<LAYOUT_CAPTURE> LAYOUT_CAPTURE_VEC;         //!< Utility typedef to make programming a little easier.

//! Information used in the creation of a partition header.
typedef struct LAYOUT_PARTITION
{
    //! What type of partition are we laying down.
    uint32_t signature;

    //! How many sectors to pre-assign to the partition. 0 means to use
    //! all the available space.
    uint64_t    part_size;

    //! The minimum frame size allowed in the partition. 0 means there
    //! is no minimum.
    uint32_t    min_frame_size;

    //! Capture header information.
    LAYOUT_CAPTURE_VEC cap_info;
} LAYOUT_PARTITION;
typedef std::vector<LAYOUT_PARTITION> LAYOUT_PARTITION_VEC;     //!< Utility typedef to make programming a little easier.

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

/*!
 * @brief The Fastec Filesystem base class.
 */
class FiFilesystem
{
private:
    static std::unique_ptr<FiFilesystem> instance;
    static std::mutex instance_mutex;

private:
    FiFilesystem();

public:
    virtual ~FiFilesystem();

    FiFilesystem(FiFilesystem& other) = delete;         // Not cloneable.
    void operator=(const FiFilesystem&) = delete;       // Not assignable.

    bool analyze(const std::string& dev_path);
    bool appendCapture(const std::string& dev_path, uint64_t part_lba, const LAYOUT_CAPTURE_VEC& layout);
    bool appendPartition(const std::string& dev_path, const LAYOUT_PARTITION_VEC& layout);
    bool clearCache(const std::string& dev_path);
    bool deleteCapture(const std::string& dev_path, uint64_t part_lba, uint64_t capture_lba);
    bool deletePartition(const std::string& dev_path, uint64_t part_lba);
    bool format(const std::string& dev_path, const std::string& volume_label, uint64_t num_sectors, const LAYOUT_PARTITION_VEC& layout);
    bool getCaptureBK(const std::string& dev_path, uint64_t capture_lba, std::vector<uint8_t>& bk_buffer);
    bool getCaptureCT(const std::string& dev_path, uint64_t capture_lba, std::vector<uint8_t>& ct_buffer);
    bool getCaptureGT(const std::string& dev_path, uint64_t capture_lba, std::vector<uint8_t>& gt_buffer);
    bool getCapturePM(const std::string& dev_path, uint64_t capture_lba, std::vector<uint8_t>& pm_buffer);
    bool getDriveInfo(const std::string& dev_path, PARTITION_INFO_VEC& part_info, uint64_t& drive_space, std::string& volume_label);
    bool mergeCaptures(const std::string& dev_path, uint64_t part_lba, const std::vector<uint64_t>& merge_lbas);
    bool rename(const std::string& dev_path, const std::string& volume_label);
    bool resizeCapture(const std::string& dev_path, uint64_t capture_lba, uint64_t resize_amount);
    bool resizePartition(const std::string& dev_path, uint64_t part_lba, uint64_t resize_amount);
    bool setCaptureBK(const std::string& dev_path, uint64_t capture_lba, const std::vector<uint8_t>& bk_buffer);
    bool setCaptureInfo(const std::string& dev_path, uint64_t capture_lba, const std::vector<uint8_t>& ct_buffer, const std::vector<uint8_t>& gt_buffer);
    bool setCapturePM(const std::string& dev_path, uint64_t capture_lba, const std::vector<uint8_t>& pm_buffer);
    bool tryLayout(const std::string& dev_path, uint64_t num_sectors, const LAYOUT_PARTITION_VEC& layout, PARTITION_INFO_VEC& part_info);
    bool updateCapture(const std::string& dev_path, const CAPTURE_DATA& capture_info);
    bool writePartition(const std::string& dev_path, uint64_t num_sectors, const PARTITION_INFO_VEC& part_info);

    static int getReserved(uint32_t type);

    static FiFilesystem *getInstance();
};

} /* namespace FAS_FILESYSTEM */

#endif // FAS_FILESYSTEM_H_

