/*!
 * @file fas_internal.h
 * @brief The main include file for the Fastec Internal Proprietary library.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FASTEC_INTERNAL_H_
#define FASTEC_INTERNAL_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <string>
#include <stdint.h>

namespace FAS_INTERNAL
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

//! Fastec Internal Proprietary library's error codes.
typedef enum FIP_STATUS
{
    FIP_OK,                                 //!< Everything is fine.
    STREAM_EMPTY,                           //!< Video stream object is not constructed and waiting to be created.
    STREAM_BUILT,                           //!< Video stream object is only constructed and not ready to be used.
    STREAM_READY,                           //!< Video stream object is prepared and ready for use.
    STREAM_COMPLETED,                       //!< Video stream object has completed processing.
    STREAM_CLOSED,                          //!< Video stream object has closed and is no longer available for use.
    STREAM_INCOMPLETE,                      //!< Video stream object is incomplete but asked to close.
    STREAM_TYPE_ERROR,                      //!< Video stream object type is incompatible with request.
    PLACEHOLDER_WAIT,                       //!< Video stream object is waiting for placeholder frames.
    INVALID_CODEC,                          //!< Video stream object does not support codec type requested.
    INVALID_HANDLE,                         //!< Stream handle is invalid.
    BAD_RANGE,                              //!< Video stream object received a bad frame range for stream.
    NO_IO_STREAM,                           //!< Video stream object no I/O stream exists for file reads/writes.
    NO_STREAM_IMAGE,                        //!< Video stream object no FiImage handler for image management.
    NO_FILE_FOUND,                          //!< Video stream object received file information for non-existant file.
    REQUEST_DENIED,                         //!< Video stream object unable to service request due to stream status.
    STREAM_REENTRY_ERROR,                   //!< Video stream object received a frame when it was still processing one.
    ALLOCATION_FAILED,                      //!< Video stream object was unable to allocate internal objects.
    STREAM_READ_ERROR,                      //!< Video stream object was unable to read from the stream.
    STREAM_EOF,                             //!< Video stream object reached eof in the stream.
    STREAM_SETUP_ERROR,                     //!< There was an error in the setup of the stream.
    CODEC_SETUP_ERROR,                      //!< Video stream object received codec setup information that was invalid.
    CODEC_ERROR,                            //!< Video stream object received codec error while creating video.
    EXCEPTION_CAUGHT                        //!< Video stream object received unhandled exception from system.
} FIP_STATUS;

//! Fastec Internal Proprietary enumeration of the various binning options supported.
typedef enum FIP_SENSOR_BINNING
{
    NO_BINNING,                             //!< There was no binning applied.
    BINNING_2X,                             //!< 2X binning was applied.
    BINNING_4X                              //!< 4X binning was applied.
} FIP_SENSOR_BINNING;

//! Fastec Internal Proprietary enumeration of the various subsampling options supported.
typedef enum FIP_SENSOR_SUBSAMPLE
{
    NO_SUBSAMPLING,                         //!< There was no subsampling applied.
    SUBSAMPLE_2X,                           //!< 2X subsampling was applied.
    SUBSAMPLE_4X                            //!< 4X subsampling was applied.
} FIP_SENSOR_SUBSAMPLE;


//--------------------------------------------------------------------
// Type Definitions
//--------------------------------------------------------------------

//! @brief This structure contains the information necessary for configuring
//!        the Fastec Internal Proprietary stream engine.
typedef struct FIP_STREAM_CONFIGURATION
{
    std::string manufacturer;               //!< Camera manufacturer.
    std::string cameraModel;                //!< The camera model.
    std::string cameraVersion;              //!< The version of the camera used. Perhaps the software version.
    std::string macAddress;                 //!< Standard representation of the camera's MAC addres, "xx:xx:xx:xx:xx:xx".
    std::string userComment;                //!< User supplied, free-form string.

    uint32_t roiX;                          //!< The image origin (upper-left) X-coordinate.
    uint32_t roiY;                          //!< The image origin (upper-left) Y-coordinate.
    uint32_t width;                         //!< The width of the images, in pixels.
    uint32_t height;                        //!< The height of the images, in pixels.
    int frameRate;                          //!< The frame rate of the video (frames per second).
    int exposure;                           //!< The exposure (shutter speed), in microseconds.
    int bitsPerPixel;                       //!< The number of bits-per-pixel in the images.
    uint32_t frameCount;                    //!< The number of frames in the stream.

    uint32_t sensorType;                    //!< The Fastec sensor ID.
    FAS_INTERNAL::FIP_SENSOR_BINNING sensorBinningOpt;      //!< The binning mode, if any, that was used.
    FAS_INTERNAL::FIP_SENSOR_SUBSAMPLE sensorSubSampleOpt;  //!< The subsampling mode, if any, that was used.

    std::string lensManufacturer;           //!< The lens manufacture.
    std::string lensModel;                  //!< The model of the lens.
    int focalLength;                        //!< The focal length (zoom) used at the time of the capture, in millimeters.
    uint32_t aperture;                      //!< The aperture (F-stop) used at the time of the capture. In 10ths, so an F-stop of "3.5" would be "35".

    uint64_t timeStartOfStreamSeconds;      //!< The seconds portion of the time for the first frame at the start of the stream, in seconds since epoch (Linux time).
    uint32_t timeStartOfStreamMicrosec;     //!< The microseconds portion of the time for the first frame at the start of the stream.
} FIP_STREAM_CONFIGURATION;

//! Create an opaque type for our packed stream.
typedef uint32_t    FIP_HANDLE;

} // End of FAS_INTERNAL

//--------------------------------------------------------------------
// Function Prototypes
//--------------------------------------------------------------------

/*! @brief Calculate how many bytes the packed data for an image width x height
 *         pixels (bitsPerPixels deep) would take.
 *
 *  @param width        The image width, in pixels.
 *  @param height       The height of the image.
 *  @param bitsPerPixel The number of bits per pixel: 8, 10, or 12.
 *
 *  @return The number of bytes in the packed frame.
 */
extern "C" uint32_t fiGetPackedSize(uint32_t width, uint32_t height, int bitsPerPixel);

/*!
 * @brief Create the packed streamer object.
 *
 * @param streamName    The fully qualified path and name for the packed streamer's main file.
 * @param configInfo    The configuration information for the data stream.
 * @return An opaque handle uniquely referencing the packed streamer.
 */
extern "C" FAS_INTERNAL::FIP_HANDLE fiCreatePackStreamer(const std::string& streamName, const FAS_INTERNAL::FIP_STREAM_CONFIGURATION& configInfo);

/*!
 * @brief Close the packed streamer along with any open files.
 *
 * @param streamHandle  The opaque handle uniquely referencing the packed streamer.
 * @return FIP_STATUS::STREAM_CLOSED on success, an error status otherwise.
 */
extern "C" FAS_INTERNAL::FIP_STATUS fiClosePackStreamer(FAS_INTERNAL::FIP_HANDLE streamHandle);

/*!
 * @brief Check to see if the packed streamer has finished operations.
 *
 * @param streamHandle  The opaque handle uniquely referencing the packed streamer.
 * @return FIP_STATUS::STREAM_COMPLETED, or one of the other status codes.
 */
extern "C" FAS_INTERNAL::FIP_STATUS fiIsPackStreamerFinished(FAS_INTERNAL::FIP_HANDLE streamHandle);

/*!
 * @brief Add a TIFF image to the stream referenced by streamHandle.
 *
 * This will convert the TIFF data to an appropriate packed image.
 *
 * @param streamHandle  The opaque handle uniquely referencing the packed streamer.
 * @param imagePtr      Pointer to the TIFF data.
 * @return FIP_STATUS::STREAM_READY, or one of the other status codes.
 */
extern "C" FAS_INTERNAL::FIP_STATUS fiAddTiffImageData(FAS_INTERNAL::FIP_HANDLE streamHandle, const uint8_t *imagePtr);

/*!
 * @brief Add a placeholder to the stream referenced by streamHandle.
 *
 * @param streamHandle  The opaque handle uniquely referencing the packed streamer.
 * @return FIP_STATUS::STREAM_READY, or one of the other status codes.
 */
extern "C" FAS_INTERNAL::FIP_STATUS fiAddPlaceholder(FAS_INTERNAL::FIP_HANDLE streamHandle);

/*!
 * @brief Typically used to replace a placeholder with actual TIFF data.
 *
 * This will convert the TIFF data to an appropriate packed image.
 *
 * @param streamHandle  The opaque handle uniquely referencing the packed streamer.
 * @param seqNumber     The sequence number of the data to be replaced.
 * @param imagePtr      Pointer to the TIFF data.
 * @return FIP_STATUS::STREAM_READY, or one of the other status codes.
 */
extern "C" FAS_INTERNAL::FIP_STATUS fiReplaceFrame(FAS_INTERNAL::FIP_HANDLE streamHandle, uint32_t seqNumber, const uint8_t *imagePtr);

/*!
 * @brief Retrieve the statistics for the save.
 *
 * @param [in]streamHandle  The opaque handle uniquely referencing the packed streamer.
 * @param [out]frameCount   The number of frames to be saved.
 * @param [out]lastFrame    The last frame (sequence number) to be saved.
 * @param [out]encodedFrames The total number of frames encoded.
 * @return FIP_STATUS::FIP_OK, or one of the other status codes.
 */
extern "C" FAS_INTERNAL::FIP_STATUS fiGetPackStatistics(FAS_INTERNAL::FIP_HANDLE streamHandle, uint32_t& frameCount, uint32_t& lastFrame, uint32_t& encodedFrames);

#endif /* !FASTEC_INTERNAL_H_ */
