/*!
 * @file fas_thread.h
 * @brief The main include file for the Fastec Thread library.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FASTEC_THREAD_H_
#define FASTEC_THREAD_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>

#include <condition_variable>
#include <memory>
#include <mutex>
#include <queue>
#include <thread>

namespace FAS_THREAD
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------
typedef void (*THREAD_STARTUP)(void *t);

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

class Thread
{
public:
    Thread(bool is_detached = false);
    virtual ~Thread();

    virtual void join();

    virtual void start(THREAD_STARTUP threadFn = nullptr);
    virtual void stop();

    virtual int sleep(const struct timespec *delay_ptr);
    virtual int sleep(int sleep_sec);
    virtual int sleep_ms(uint32_t delay_ms);

    virtual int wait_for_done(uint32_t delay_ms = 0);
    virtual int wait_for_ready(uint32_t delay_ms = 0);

protected:
    volatile bool stop_request;     //! True to request the thread to stop.

    virtual void notify_done()      { wait_for_done_cond.notify_one(); }
    virtual void notify_ready()     { wait_for_ready_cond.notify_one(); }

    //! The run method MUST be provided by the deriving object.
    virtual void run() = 0;

private:
    // The underlying thread object.
    std::unique_ptr<std::thread> thread_ptr;

    // Our sleep control variables.
    std::mutex sleep_mutex;
    std::condition_variable sleep_cond;

    std::mutex wait_for_done_mutex;
    std::condition_variable wait_for_done_cond;

    std::mutex wait_for_ready_mutex;
    std::condition_variable wait_for_ready_cond;

    //! True if thread created in detached state.
    bool detached;

    // Thread entry function, calls run().
    static void start_thread(void *t);
};

} /* namespace FAS_THREAD */

// Include our Fastec thread sub-headers, so there is one include file required for the library.
// Note that we do the include here because of dependency ordering issues.
#include "fi_exception.h"
#include "fi_message.h"
#include "fi_msg_thread.h"
#include "fi_note.h"
#include "fi_sync_message.h"
#include "fi_ticker.h"

#endif /* FASTEC_THREAD_H_ */
