/*!
 * @file fas_video.h
 * @brief The main include file for the Fastec Video library.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FASTEC_VIDEO_H_
#define FASTEC_VIDEO_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
// #define INCLUDE_RDI_PACKED 1

#include <stdint.h>
#include <memory>
#include <vector>

#ifdef _WIN32
#include "gige_msg.h"
#else
#include <fastec/gige_msg.h>
#endif

namespace FAS_VIDEO
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

//! While working through the streamers I want to be able to turn on
//! verbose type of messages - this will allow me to do that.
//#define USE_VERBOSE_INFO_DEBUG

//! Create a size constant to provide sanity checks related to a FAT32
//! file system.  This is legacy code being used by camera_explore.cpp
//! and save_video.cpp in FasMotion -
const uint64_t FI_VIDEO_MAX_FAT32_SIZE =  0xffffffffLL;

//! Create a constant for a default frame rate value.
const uint32_t FI_DEFAULT_FRAME_RATE = 25;

//! @defgroup FIVID_constants Common Constant Values
//! @brief Common constant values shared across the library.
//! @{

//! @name Define the "FourCC" values used by the library.
//! @anchor fv_fourcc
//!
//! This is a blatant rip-off of the libva values.
//! @sa https://github.com/intel/libva/blob/master/va/va.h.
//! @{

//! ARGB: packed 8-bit RGBA.
//! Four bytes per pixel: alpha, red, green, blue.
const uint32_t FOURCC_ARGB  = 0x42475241;

//! RGB: Basic Windows bitmap format, 24-bit/pixel.
const uint32_t FOURCC_RGB   = 0x32424752;

//! @brief BYR1: Raw Bayer data with 8-bits per sample.
const uint32_t FOURCC_BYR1  = 0x31525942;

//! @brief BYR2: Raw Bayer data with 12-bits per sample stored
//!    in 16-bit words.
const uint32_t FOURCC_BYR2  = 0x32525942;

//! @brief NV12: two-plane 8-bit YUV 4:2:0.
//! The first plane contains Y, the second plane contains U and V in pairs of bytes.
const uint32_t FOURCC_NV12  = 0x3231564E;

//! @brief Y8: 8-bit greyscale.
//! Only a single sample, 8 bit Y plane for monochrome images.
const uint32_t FOURCC_Y8    = 0x20203859;

//! @brief Y16: 16-bit greyscale.
//! Only a single sample, 16 bit Y plane for monochrome images.
const uint32_t FOURCC_Y16   = 0x20363159;

//! @brief NONE: There is no format information available.
//! Construction of empty frame data will have this FOURCC code.
const uint32_t FOURCC_NONE   = 0x0;

//! @}

//! Defines the codecs we support.
typedef enum
{
    //! @name Video CODECs we support.
    //! @{
    FI_MJPEG_JPEG_CODEC,        //!< Motion JPEG with a collection of JPEGs.
    FI_MJPEG_DIB_CODEC,         //!< Motion JPEG with a collection of DIBs.
    FI_H264_CODEC,              //!< H.264 MP4 software video codec.
    FI_RDI_PACKED_CODEC,        //!< RDI Packed Image collection.
    //! @}

    //! @name Image CODECs we support.
    //! @{
    FI_BMP_STACK_CODEC,         //!< Standard BMP color or greyscale.
    FI_DNG_STACK_CODEC,         //!< Adobe DNG color images [RAW format].
    FI_JPG_STACK_CODEC,         //!< JPEG images.
    FI_TIFF_STACK_CODEC,        //!< TIFF images, color or greyscale. May be RAW.
    FI_PLANAR_STACK_CODEC,      //!< Fastec planar images.
    //! @}

    FI_INVALID_CODEC            //!< Oops! This one isn't allowed!
} VIDEO_CODEC;

typedef std::vector<uint8_t>    HEADER_DATA;
typedef std::vector<uint8_t>    X_CFG_METADATA_VECTOR;

#pragma pack(1)
//! Defines data structure used by metadata for gain information.
typedef struct VIDEO_GAINS
{
    uint32_t    red;
    uint32_t    green;
    uint32_t    blue;
} VIDEO_GAINS;
#pragma pack()

//! Define the compression speed settings for the H.264 encoder.
//! The spped is stored in range order of slowest compression time
//! to fastest compression time.  The slower the compression time
//! the better the compression and lower the filesize.  The faster
//! the compression time the lower the compression and the higher
//! the filesize.  Broadly speaking the faster the compression time
//! the higher the playback bitrates as well.
typedef enum
{
    H264_SLOWER = 0,
    H264_MEDIUM,
    H264_FAST,
    H264_VERY_FAST,
    H264_END_OF_SPEEDS, // Keep this at the end of the list.
} H264_SPEEDS;

//! Define a default speed setting for the H.264 encoder.
const FAS_VIDEO::H264_SPEEDS DEFAULT_H264_SPEED = FAS_VIDEO::H264_SPEEDS::H264_MEDIUM;

//! Define a default quality setting for the H.264 encoder.
//! The default to use is 70%.
const uint32_t DEFAULT_H264_QUALITY_PCT = 70;

//! Define a minimum quality setting for the H.264 of 20%.
const uint32_t MINIMUM_H264_QUALITY_PCT = 20;

//! The Keyframe Interval sets how often the keyframes will be forced
//! in the MP4 video stream.  The user right now can select a keyframe
//! be placed in everyframe of the stream (helps with video scrubbing)
//! or the default will be set here at every 60 frames.  The setting of
//! 60 seems to be acceptable in terms of scrubbing response.  In the
//! future we may make this value setable by the user.  The Quicktime
//! default for MP4 video is 60 frames as well as a historical note.
const uint32_t DEFAULT_H264_KEYFRAME_INTERVAL = 60;

//! Define the bits per pixel for a stream of TIFF raw images.
//! TIFF raw images can be streamed in 8-BPP format or 16-BPP format. When
//! receiving raw TIFF images in 16-BPP only 10 or 12 bits of the 16 will
//! be true image data. This is how the stream writer will know what to
//! expect when raw TIFF images are read as input for a transcode operation.
//!
//! Should the stream writer receive FRAME_BPP_UNUSED then it knows that raw
//! TIFF images are not expected in the input stream.
//!
//! Right now only RDI packed stream writers take raw TIFF images as input.
typedef enum
{
    FRAME_8_BPP = 0,
    FRAME_10_BPP,
    FRAME_12_BPP,
    FRAME_BPP_UNUSED
} FRAME_BPP;

//! @}  // End of FIVID_constants

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

// Still having issues when these circular include files are used here.
// Just using a forward class declaration to get around these problems.
class FiVideoStream;
class FiImage;

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------
class FiVideo
{
private:
    FiVideo();

public:
    virtual ~FiVideo();

    static std::unique_ptr<FAS_VIDEO::HEADER_DATA> CreateImageHeader(FAS_VIDEO::VIDEO_CODEC type, uint32_t fourcc, pm_metadata *meta);

    static std::unique_ptr<FAS_VIDEO::FiVideoStream> CreateAVIVideoStreamReader(const std::string& videoName);

    static std::unique_ptr<FAS_VIDEO::FiVideoStream> CreateAVIVideoStreamWriter(const std::string& videoName,
                                                                                uint32_t frameRate,
                                                                                uint32_t frameWidth,
                                                                                uint32_t frameHeight,
                                                                                uint32_t frameCount);

    static std::unique_ptr<FAS_VIDEO::FiVideoStream> CreateMP4VideoStreamWriter(const std::string& videoName,
                                                                                uint32_t frameRate,
                                                                                uint32_t frameWidth,
                                                                                uint32_t frameHeight,
                                                                                uint32_t frameCount,
                                                                                uint32_t h264QualityPct = FAS_VIDEO::DEFAULT_H264_QUALITY_PCT,
                                                                                FAS_VIDEO::H264_SPEEDS h264Speed = FAS_VIDEO::H264_SPEEDS::H264_MEDIUM,
                                                                                uint32_t h264KeyInterval = FAS_VIDEO::DEFAULT_H264_KEYFRAME_INTERVAL);

#ifdef INCLUDE_RDI_PACKED
    static std::unique_ptr<FAS_VIDEO::FiVideoStream> CreateRDIPackedVideoStreamWriter(const std::string& videoName,
            uint32_t frameRate,
            uint32_t frameWidth,
            uint32_t frameHeight,
            uint32_t frameCount,
            const pm_metadata_type *pm_ptr);
#endif /* INCLUDE_RDI_PACKED */

    static bool InitializeLogger(std::string& logDirectory, int maxLogsBeforeRotate);

    static void H264SpeedSettingToText(std::string& speedText, FAS_VIDEO::H264_SPEEDS speedSetting);

    static int32_t CreateMetadataString(std::string& metadataStringRef,
                                        const void *pmMetadataPtr, uint32_t pmMetadataSize,
                                        uint32_t startFrame, uint32_t endFrame,
                                        const FAS_VIDEO::VIDEO_GAINS *gainsPtr);

    static int32_t CreateX_CFG_MetadataString(std::unique_ptr<FAS_VIDEO::X_CFG_METADATA_VECTOR>& xcfgMetaDataInBytes,
                                              const void *pmMetadataPtr, uint32_t pmMetadataSize,
                                              uint32_t startFrame, uint32_t endFrame);

    static const std::string GetFourCCReadableString(uint32_t fourccToTranslate);

    static const std::string& GetIncompleteVideoFileString();
};

} // End of FAS_VIDEO


#endif /* !FASTEC_VIDEO_H_ */
