/*!
 * @file ffmpeg_filter.h
 * @brief Interface for FFmpeg filtering object.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FFMPEG_FILTER_H_
#define FFMPEG_FILTER_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#ifdef  __cplusplus
extern "C" {
#endif

#include "libavformat/avformat.h"
#include "libavfilter/buffersink.h"
#include "libavfilter/buffersrc.h"
#include "libavformat/avio.h"
#include "libavcodec/avcodec.h"
#include "libavutil/file.h"
#include "libavutil/error.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"

#ifdef  __cplusplus
}
#endif

#include "fas_video.h"

using namespace std;
using namespace FAS_VIDEO;

namespace FI_FFMPEG
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Declarations
//--------------------------------------------------------------------

//! Create a class to handle managment of the filter information.
class FFmpegFilters
{
public:
    FFmpegFilters() :
        buffersink_ctx(nullptr),
        buffersrc_ctx(nullptr),
        filter_graph(nullptr),
        FFmpegStatus(0),
        filtersPrepared(false)
    {
        // not doing anything yet.
    }
    virtual ~FFmpegFilters();


    //! @brief Setup the FFmpeg filters.
    //!
    //! Filters operate with a buffer src and buffer sink to manipulate frame data.
    //!
    //! @return boolean true if filters are initialized properly or false if error occurred.
    //!
    virtual bool InitializeFilters(AVRational& avrAspectRatio,
                                   AVRational& avrTimeBase,
                                   uint32_t frameWidth,
                                   uint32_t frameHeight,
                                   AVPixelFormat inFormat,
                                   AVPixelFormat outFormat);

    //! @brief Return the state of filter preparation.
    //!
    //! @return true if the filter is prepared for use.
    //!
    bool is_filter_prepared();

    //! @brief Return the AVFilterContext pointer for prepared filter.
    //!
    //! @return the AVFilterContext pointer manging source data or nullptr if not prepared.
    //!
    AVFilterContext *get_buffer_src_ctx();

    //! @brief Return the AVFilterContext pointer for prepared filter.
    //!
    //! @return the AVFilterContext pointer managing sink (dest) data or nullptr if not prepared.
    //!
    AVFilterContext *get_buffer_sink_ctx();

    //! @brief Return the AVFilterGraph pointer for prepared filter.
    //!
    //! @return the AVFilterGraph pointer (TBD not sure how filter graphs are used) we have or nullptr if not prepared.
    //!
    AVFilterGraph *get_filter_graph();


private:
    //! FFmpeg specific data structures necessary for filtering frame data.
    AVFilterContext *buffersink_ctx;
    AVFilterContext *buffersrc_ctx;
    AVFilterGraph *filter_graph;

    int FFmpegStatus;      //! Holds an FFmpeg status code (0) means no errors occurred.

    bool filtersPrepared;  //! True if ready to encoder, false if not.
};

} /* End of namespace FI_FFMPEG */

#endif /* !FFMPEG_FILTER_H_ */
