/*!
 * @file fi_bmp.h
 * @brief Definitions and declarations for Fastec BMP images.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FIVID_FI_BMP_
#define FIVID_FI_BMP_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <memory>
#include <vector>

#include "fas_video.h"
#include "fi_image.h"

namespace FI_BMP
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! The BMP magic value.
extern const char *BMP_MAGIC;

//! 256 colors, 32 bits per entry (RGBA).
const int DIB_COLOR_TABLE_SIZE      = (32 * 256);

//! Defines the row alignment requirement for 24 BPP and 8 BPP BMP images.
const uint32_t ROW_ALIGNMENT        = 4;

#pragma pack(1)
//! The "Device Independent Bitmap" portion of BMP header.
typedef struct DIB_HEADER
{
    //! Size of this structure.
    uint32_t data_header_size;

    //! Image width.
    int32_t width;

    //! Image Height.
    int32_t height;

    //! Number of color planes, MUST be set to 1.
    uint16_t color_planes;

    //! Number of bits per pixel.
    uint16_t depth;

    //! Compression method:
    //!     0   None
    //!     1   RLE 8-bit/pixel, only used with 8-bit/pixel bitmaps
    //!     2   RLE 4-bit/pixel, only used with 4-bit/pixel bitmaps
    //!     3   Huffman 1D compression
    //!     4   RLE-24 compression
    //!     5   PNG, contains a PNG image
    uint32_t compression;

    //! The size of the bitmap data (following the header).
    uint32_t data_size;

    //! Width resolution, set to 0 to indicate no preference.
    uint32_t width_resolution;

    //! Height resolution, set to 0 to indicate no preference.
    uint32_t height_resolution;

    //! The number of colors in the color palette, set to 0 to allow default.
    uint32_t actual_colors_used;

    //! The number of important colors used, set to 0 to indicate every color is important.
    uint32_t important_colors_used;
} DIB_HEADER;
#pragma pack(0)

#pragma pack(1)
//! The generic BMP header.
typedef struct BMP_HEADER
{
    // uint16_t bmp_id - the BMP "magic" signature. To avoid misaligning everything,
    //           we'll leave this out and copy it in later.
    uint32_t length;        //!< Total length of the bitmap file (header+data).
    uint8_t reserved[4];    //!< Reserved, set to 0
    uint32_t data_offset;   //!< Offset to the start of bitmap data from the beginning.

    DIB_HEADER dib;     //!< Contains "Device Independent Bitmap" portion of header.
} BMP_HEADER;
#pragma pack(0)

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

class FiBMP : public FAS_VIDEO::FiImage
{
public:
    FiBMP();
    virtual ~FiBMP() {}

    virtual int CreateHeader(uint32_t fourcc, const pm_metadata *meta) override;
    std::unique_ptr<FAS_VIDEO::HEADER_DATA> GetHeader() override;

    virtual const uint8_t *GetAttachedFrameData(uint32_t& attachedFrameSize,
                                                FAS_VIDEO::FRAME_PLANAR_DATA& planarDataAttached,
                                                FAS_VIDEO::FRAME_ORIENTATION frameOrientation = FAS_VIDEO::FRAME_ORIENTATION::TOP_DOWN) override;

    //! @name Public Property Accessor Functions.
    //!
    //! These are the public functions used to access the properties of this object.
    //! @{

    virtual FAS_VIDEO::IMAGE_FORMAT GetImageFormat() override;
    virtual uint32_t GetImageWidth() override;
    virtual uint32_t GetImageHeight() override;
    virtual uint32_t GetBitsPerPixel() override;
    virtual uint32_t GetRowAlign() override;
    virtual uint32_t GetFullFramesize() override;
    virtual uint32_t GetHeaderSize();

    //! @}

protected:
    virtual bool VerifyImageData(const uint8_t *imageDataToVerify, uint32_t imageDataSize) override;
    virtual bool GetAttachedFramePlanarData(FAS_VIDEO::FRAME_PLANAR_DATA& planarDataToSet) override;

    void init_color_table();
    bool isBmpHdr(const char *imageDataPtr);
    bool BMPHeaderExists();
    virtual bool CopyBMPHeaderData(const uint8_t *imageDataToVerify, uint32_t imageDataSize);

    bool createdHeader;            //!< True if this is an external header, false if it was from an image.
    int hdr_size;                   //!< The total number of bytes in the "total" header.
    BMP_HEADER header;              //!< Just the BMP header.
    std::vector<char> color_table;  //!< The color table for greyscale images.

    //! Attached Image Frame Data.
    std::unique_ptr<FAS_VIDEO::FiFrame> frameData;

    //! Attached Image Frame Planar Data
    FAS_VIDEO::FRAME_PLANAR_DATA framePlaneData;
};

} // End of FI_BMP

#endif  /* FIVID_FI_BMP_ */

