/*!
 * @file fi_filesys.h
 * @brief This is the header file for the Fastec Imaging file system.
 *
 * Copyright 2013-2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */
#ifndef _FI_FILESYS_H_
#define _FI_FILESYS_H_

namespace Fastec
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Frame metadata word offsets.
constexpr int FMETA_TIME_SECONDS	= 0;
constexpr int FMETA_TIME_USECS		= 1;
constexpr int FMETA_MARKERS		= 2;
constexpr int FMETA_HW_NUM 		= 3;
constexpr int FMETA_FIFO_LEVELS 	= 4;
constexpr int FMETA_RECORDING_ID	= 5;

//! Mask to extract microseconds from frame metadata word that contains some other flags.
constexpr uint32_t FMETA_TIME_USEC_MASK = 0xfffff;

//! Signatures for file system headers
constexpr char *DRIVE_HD_SIG		= "LRDV";
constexpr char *DPART_HD_SIG		= "PART";
constexpr char *CAPTURE_HD_SIG		= "CAPT";

//! States of a CAPTURE header
enum CAPTUREHD_STATE {
    CAPTUREHD_ALLOCATED,
    CAPTUREHD_READY,
    CAPTUREHD_CAPTURING,
    CAPTUREHD_COMPLETE,
    CAPTUREHD_APPENDING
};

//! Flags in the CAPTURE header
constexpr int CAPHD_FLAG_RECONSTRUCTED	= 0x01;	//!< Reconstructed start+end after a crash.
constexpr int CAPHD_FLAG_APPENDMODE	= 0x02;	//!< ROC or BROC mode recording.

//! How many drive partitions we can store on a drive.
constexpr int MAX_LTR_DPARTS	= 40;

#pragma pack(1)
typedef struct FI_SIG
{
    uint8_t 	signature[4];	// Signature for this type of header.
    uint16_t	size;		// Size in bytes of data beyond the header.
    uint16_t	checksum;	// Checksum covering size bytes.
} FI_SIG;

typedef struct DRIVE_HD
{
    uint64_t	drive_size;	// The drive size on which this format was based.
    uint16_t	alloc_unit;	// The size of allocation units in sectors.
    // This is for info and future use, since we code with the defined
    // constant SECTORS_PER_PAGE.
    uint16_t	placeholder[3];	// Reserved for future use, set to zero.
    uint64_t	dpart_lba[ MAX_LTR_DPARTS+1 ]; // Array of drive partition LBAs, ending with zero.
} DRIVE_HD;

// "Standard" drive partition header for holding captures. If we have different types of partitions
// later, e.g. one to hold CAP files, they can have a different signature.
// A partition can have multiple child captures linked together, which all share the common
// partition image and frame metadata regions.
// The minimum frame size we used to decide how much frame metadata to allocate is probably
// unnecessary since all we really care about is whether there is space left to support the next
// capture, but it may be useful while getting this running.
typedef struct DPART_HD
{
    uint64_t fmeta_lba;		// Preallocated frame metadata region shared by all the captures
    // within this partition.
    uint64_t fmeta_sectors;	// Size of frame metadata region.
    uint64_t image_lba;		// Location of the start of image data corresponding to fmeta_lba.
    uint64_t image_sectors;	// Image area size on which the fmeta allocation was based. Appended
    // captures will use this to determine remaining usable space.
    uint64_t capturehd_lba;	// Location of the first capture header for this partition.
    // Capture metadata must fit between capturehd_lba and image_lba.
    uint32_t min_frame_size;	// Minimum frame size on which the fmeta allocation was based.
    // If not permitting append, this can be the actual frame size.
    uint8_t  reserved_sectors;	// Always 1 for now.
    uint8_t  placeholder[3];	// Reserved for future use.
} DPART_HD;

// Describes a single capture.
typedef struct CAPTURE_HD
{
    // We need the linkage to the parent and siblings to traverse the file system.
    uint64_t dpart_lba;		// Header location of drive partition to which this capture belongs.
    uint64_t previous_lba;	// Location of previous capture header.
    uint64_t next_lba;		// Location of next capture header.

    // This describes where the image data and frame metadata are on the drive.
    // They are set if the header is in the ALLOCATED state or later.
    uint64_t fmeta_lba;		// Start of per-frame metadata for this capture.
    uint64_t fmeta_sectors;	// Number of sectors allocated for frame metadata for this capture.
    uint64_t image_lba;		// Start of image data for this capture.
    uint64_t image_sectors;	// Number of sectors allocated for image data.

    // These are other metadata structures for the capture.
    // These are set if the capture header is in the READY state or later.
    uint32_t bk_bytes;		// Size of bookmark area.
    uint32_t bk_offset;		// Byte offset from start of capture header sector to bookmark area.
    uint32_t pm_bytes;		// Size of this capture's memory partition metadata.
    uint32_t pm_offset;		// Byte offset from start of capture header sector to PM area.
    uint32_t ct_bytes;		// Size of Correction Table area.
    uint32_t ct_offset;		// Byte offset from start of capture header sector to CT area.
    uint32_t reserved_sectors;	// Total capture reserved bytes including this header.

    // Some of this is redundant with the memory partition metadata. Having it here lets us
    // write the completion information as a single sector once the capture is done.
    // We update it at the end of a capture then when memory partition metadata is loaded, we
    // will update the pm_metadata in memory based on what is here.
    // We also need some of this information to determine what space can be freed up if we want
    // to break a capture region into a used area and a free area after a capture stops.

    // These are set for a header that is in the CAPTURING state or later and for the
    // APPENDING state or later.
    uint32_t recording_id;	// ID for this recording.
    uint32_t frame_size;	// Frame size in bytes.
    uint32_t frame_sectors;	// Frame size in sectors.
    uint32_t frame_count;	// Number of frames allocated.

    // These are set for a header that is in the COMPLETED state.
    uint32_t num_frames;	// Number of frames actually captured, once capture is done.
    uint32_t first_frame_num;	// Oldest frame, number relative to first frame of the capture area.
    uint32_t trigger_frame_num;	// Trigger frame, number is relative to first_frame_num.
    uint32_t trigger_time_s;	// Trigger time seconds.
    uint32_t trigger_time_ns;	// Trigger time ns past the second (but only usec precision).

    uint8_t state;		// State of this capture: ALLOCATED, READY, CAPTURING, COMPLETED
    // ALLOCATED - frame metadata and image allocations are defined
    // READY -     params describe a capture but recording is not
    //	       started.  Shouldn't have any need for this on the
    //	       disk, but useful for dragging parameters around.
    // CAPTURING - Saved when starting a normal capture. See above.
    // APPENDING - Saved when starting a ROC/BROC capture. See above.
    // COMPLETE -  Saved when capture is complete or after post-crash
    //             recovery.

    uint8_t flags;		// Set a flag if we reconstructed the recording from the CAPTURING
    // state.

    uint8_t placeholder[2];	// Keep the size a multiple of 32-bit words.
} CAPTURE_HD;
#pragma pack(0)

} /* namespace Fastec */

#endif // _FI_FILESYS_H_

