/*!
 * @file fi_image_pkg.h
 * @brief The Fastec Image Package Class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FASTEC_IMAGE_PKG_H_
#define FASTEC_IMAGE_PKG_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>
#include <memory>
#include <vector>

namespace FAS_VIDEO
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------
//! Define a generic vector array containing 8 bits of data to represent
//! the binary blob of image data that can be attached and detached from
//! this FiImage object.
typedef std::vector<uint8_t>    IMAGE_DATA;

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

/*!
 * @brief The Fastec Imaging Image Package class
 *
 * When moving image data around in the system depending on the source of the data
 * it could be externally managed by the camera server or it could be something that
 * a FiImage derivation may manage in an IMAGE_DATA vector.  The purpose of this
 * object is to wrap image data in a single package to move through the system.  It
 * will manage and provide image data processed by the fas_video library.
 */
class FiImagePkg
{
public:
    virtual ~FiImagePkg();

    //! Factory methods for creating an image package.
    static std::unique_ptr<FAS_VIDEO::FiImagePkg> Create(const uint8_t *memDataPtr, uint32_t memDataLen);
    static std::unique_ptr<FAS_VIDEO::FiImagePkg> Create(std::unique_ptr<FAS_VIDEO::IMAGE_DATA> imageDataToOwn);

    virtual uint32_t GetImageSize();
    virtual const uint8_t *GetMemDataPtr();

protected:
    FiImagePkg();

    //! This object is non-copyable and non-copy assignable.
    FiImagePkg(const FiImagePkg& i) = delete;
    FiImagePkg& operator=(const FiImagePkg& i) = delete;

    //! Packaging up externally managed image data will reside in these member
    //! variables - this data is read-only and non-modifiable.
    const uint8_t *externalImagePtr;
    uint32_t externalImagePtrLen;

    //! Packaging up internally managed image data will reside in this member
    //! variable.  This data is managed by this package and modifiable.
    std::unique_ptr<FAS_VIDEO::IMAGE_DATA> internalImageData;
};

} // End of FAS_VIDEO

#endif // FASTEC_IMAGE_PKG_H_
