/*!
 * @file fi_msg_thread.h
 * @brief The definitions and declarations for the MsgThread class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_MSG_THREAD_H_
#define FI_MSG_THREAD_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <fas_thread.h>

namespace FAS_THREAD
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

class MsgThread : public FAS_THREAD::Thread
{
public:
    MsgThread();
    virtual ~MsgThread();

    virtual void messageAdd(FAS_THREAD::Message *message);
    virtual void run();
    virtual void start(FAS_THREAD::THREAD_STARTUP threadFn = nullptr);
    virtual bool isHalted() { return halted; }

protected:
    /*!
     * @brief Handle one message
     *
     * Derived classes must provide an implementation for this method. It does
     * the work of the MsgThread. It executes in the MsgThread's context once for
     * each message that has been placed on the MsgThread's queue.
     *
     * @param message Pointer to the message to process.
     */
    virtual void messageHandle(FAS_THREAD::Message *message) = 0;
    virtual void ownerMessage(FAS_THREAD::Message *message);

    /*!
     * @brief Convert the message request code to a string identifier.
     */
    virtual std::string reqToString(int req_code);

    //! @name MsgThread startup synchronization
    //!
    //! The base class provides these synchronization variables and
    //! methods for use by derived classes.
    //!
    //! @{

    virtual void hold();
    virtual void release();

    FAS_THREAD::Message *wait_for_message(int delay_ms = 0);

    //! A mutex to go with the condition variable.
    std::unique_ptr<std::mutex> startup_mutex;

    //! A condition variable to go with the mutex.
    std::unique_ptr<std::condition_variable> startup_condition;

    //! State variable for the thread to test.
    bool startup_ok;
    //! @}

    //! @name Job queue variables
    //! @{

    //! @brief Message queue; input jobs for thread
    std::unique_ptr<std::queue<FAS_THREAD::Message *>> queue;

    //! @brief mutex that guards message queue
    std::unique_ptr<std::mutex> msg_mutex;

    //! @brief condition variables that guards job queue
    std::unique_ptr<std::condition_variable> jobs;
    //! @}

    // Event tallies for diagnostics
    uint64_t messagesIn;
    uint64_t messagesOut;

    //! Flag to indicate the thread has been asked to stop.
    volatile bool halted;

private:
    /*!
     * @brief The default thread entry routine.
     */
    static void threadEntry(void *t)
    {
        ((MsgThread *)t)->hold();
        ((MsgThread *)t)->run();
    }
};

} /* namespace FAS_THREAD */

#endif /* FI_MSG_THREAD_H_ */
