/*!
 * @file fi_stream.h
 * @brief The stream object of the Fastec Video library.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FASTEC_STREAM_H_
#define FASTEC_STREAM_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <memory>
#include <string>
#include <vector>
#include <fstream>

#include "fas_video.h"
#include "fi_image.h"
#include "fi_frame.h"

namespace FAS_VIDEO
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

//! Default framerate value set for 25 frames per second.
const uint32_t FI_DEFAULT_VIDEO_FRAME_RATE  = 25;

//! Defines the max size of the text representation of the TS3 metadata
const uint32_t FI_METADATA_TEXT_SUGGESTED_SIZE = 8192;

//! Defines the video stream I/O type.
typedef enum
{
    //! @name Video Stream operations we support.
    //! @{
    FI_STREAM_NONE,             //!< Stream is constructed but not set for any I/O operation.
    FI_STREAM_READER,           //!< Stream is set for reading.
    FI_STREAM_WRITER,           //!< Stream is set for writing.
    //! @}
} STREAM_IO_TYPE;

//! Create object conditions of status for ffmpeg decoder and encoder objects.
typedef enum STREAM_STATUS
{
    STREAM_EMPTY,           //!< Video stream object is not constructed and waiting to be created.
    STREAM_BUILT,           //!< Video stream object is only constructed and not ready to be used.
    STREAM_READY,           //!< Video stream object is prepared and ready for use.
    STREAM_COMPLETED,       //!< Video stream object has completed processing.
    STREAM_CLOSED,          //!< Video stream object has closed and is no longer available for use.
    STREAM_INCOMPLETE,      //!< Video stream object is incomplete but asked to close.
    STREAM_TYPE_ERROR,      //!< Video stream object type is incompatible with request.
    PLACEHOLDER_WAIT,       //!< Video stream object is waiting for placeholder frames.
    INVALID_CODEC,          //!< Video stream object does not support codec type requested.
    BAD_RANGE,              //!< Video stream object received a bad frame range for stream.
    NO_IO_STREAM,           //!< Video stream object no I/O stream exists for file reads/writes.
    NO_STREAM_IMAGE,        //!< Video stream object no FiImage handler for image management.
    NO_FILE_FOUND,          //!< Video stream object received file information for non-existant file.
    REQUEST_DENIED,         //!< Video stream object unable to service request due to stream status.
    STREAM_REENTRY_ERROR,   //!< Video stream object received a frame when it was still processing one.
    ALLOCATION_FAILED,      //!< Video stream object was unable to allocate internal objects.
    STREAM_READ_ERROR,      //!< Video stream object was unable to read from the stream.
    STREAM_EOF,             //!< Video stream object reached eof in the stream.
    CODEC_SETUP_ERROR,      //!< Video stream object received codec setup information that was invalid.
    CODEC_ERROR,            //!< Video stream object received codec error while creating video.
    EXCEPTION_CAUGHT,       //!< Video stream object received unhandled exception from system.
    FFMPEG_ERR,             //!< Video stream object encountered FFmpeg Library error and can't continue.
} STREAM_STATUS;

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

class FiVideoStream
{
public:
    FiVideoStream();
    virtual ~FiVideoStream();

    // This object, for now, is non-copyable and non-copy assignable.
    FiVideoStream(const FiVideoStream& i) = delete;
    FiVideoStream& operator=(const FiVideoStream& i) = delete;

    static const char *GetStringForFiImageID(FAS_VIDEO::IMAGE_FORMAT FiImageFmtId);
    static const std::string& GetIncompleteVideoFileString();

    virtual FAS_VIDEO::STREAM_STATUS PrepareForWriting(const std::string& videoNameToCreate,
                                                       uint32_t frameRate,
                                                       uint32_t frameWidth,
                                                       uint32_t frameHeight,
                                                       uint32_t frameCount) = 0;

    virtual FAS_VIDEO::STREAM_STATUS PrepareForReading(const std::string& videoNameToOpen) = 0;

    virtual FAS_VIDEO::STREAM_STATUS AddVideoFrame(std::shared_ptr<FAS_VIDEO::FiImage> imageData) = 0;

    virtual FAS_VIDEO::STREAM_STATUS AddPlaceholderFrame() = 0;

    virtual FAS_VIDEO::STREAM_STATUS EditVideoFrame(uint32_t frame_num, std::shared_ptr<FAS_VIDEO::FiImage> imageDataFrame) = 0;

    virtual std::unique_ptr<FAS_VIDEO::FiImage> LoadVideoFrame(uint32_t frameNum) = 0;

    virtual FAS_VIDEO::STREAM_STATUS CloseStream();

    virtual bool IsStreamReady();
    virtual bool IsStreamClosed();
    virtual FAS_VIDEO::STREAM_STATUS IsStreamWriterComplete() = 0;

    virtual void LogStreamStatus();
    virtual void LogStreamStatus(std::string& logMessageStr, bool sendToFiLog = true);

    virtual uint32_t GetFrameCount();
    virtual uint32_t GetFrameRate();
    virtual const std::string GetVideoName();
    virtual uint32_t GetLastFrameAdded();
    virtual uint32_t GetFramesEncodedCount();

protected:
    STREAM_IO_TYPE streamType;              //!< Identifies the type of stream I/O supported by the stream object.
    STREAM_STATUS streamStatus;             //!< Manages the status of this video stream.
    std::string additionalStatusInfo;       //!< Some streams may want to provide additional information about streamStatus.

    uint32_t frameRate;                     //!< Number of frames to render per second of video time.
    uint32_t frameCount;                    //!< Total number of frames in the video.
    uint32_t lastFrameAdded;                //!< Frame number of last frame added to stream writer.

    uint32_t frameWidth;                    //!< Zero-based frame width is number of pixels in a row.
    uint32_t frameHeight;                   //!< Zero-based frame height is number of rows in a frame.
    uint32_t fullFrameSize;                 //!< Total count of uint8_t bytes in a full frame of data.

    std::string videoNameOfStream;          //!< Name of video for stream I/O operations.

    FAS_VIDEO::IMAGE_FORMAT firstAddedImageFormat;      //!< Ensure all image frames are same type.

    bool processingVideoFrame;              //!< Guards against re-entering AddVideoFrame before it completes.

    //! @{
    //! Chose to have two types of FiImage pointers held in object.  In case of stream writer
    //! the external source managing the data wraps it in a shared pointer.  In case of stream
    //! reader management is isolated within the object and using unique pointer.  It also makes
    //! readability clearer.  Even though this stream object is right now anyway, either a stream
    //! reader or a stream writer this separation will remain in place in case we ever combine the
    //! functionality within a single allocated object.
    std::shared_ptr<FiImage> videoFrameToWrite;         //!< Holds attached image for stream writing.
    std::unique_ptr<FiImage> curVideoFrameRead;         //!< Holds current loaded video frame from stream reader.
    //! @}

    bool attachVideoFrameToWrite(std::shared_ptr<FiImage> theImageToAttach);
};

} // End of FAS_VIDEO Namespace

#endif /* !FASTEC_STREAM_H_ */
