/*!
 * @file fi_ticker.h
 * @brief Definitions and declarations for the Ticker object.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_TICKER_H_
#define FI_TICKER_H_
#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <fas_thread.h>

#include <stdint.h>
#include <errno.h>

#include <condition_variable>
#include <memory>
#include <mutex>
#include <thread>

namespace FAS_THREAD
{
//--------------------------------------------------------------------
// Constant Definitions
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------
class Thread;

class Ticker : public FAS_THREAD::Thread
{
public:
    /*!
     * @brief Constructor for Ticker
     *
     * @param period Ticker's period, in microseconds
     */
    explicit Ticker(uint32_t period);
    virtual ~Ticker();

    /*!
     * @brief All thread objects need a run method.
     */
    virtual void run() override;

    /*!
     * @brief Change the time period
     *
     * @param period  New time period, in microseconds
     */
    virtual void setTime(uint32_t period);

    /*!
     * @brief Stop the Ticker (no more calls to tick until after another start()).
     */
    virtual void stop() override;

    /*!
     * @brief Pause/resume the tick event.
     */
    virtual void pause(bool pauseTick) { paused = pauseTick; }

    /*!
     * @brief Make the timer a single-shot timer.
     */
    virtual void setSingleShot(bool sshot) { single_shot = sshot; }

    /*!
     * @brief Pure virtual tick() method
     *
     * Concrete sub-classes must provide an implementation of the tick() method.
     * The Ticker will execute the tick() method once every 'interval' microseconds.
     */
    virtual void tick() = 0;

protected:
    uint32_t interval;
    bool paused;
    bool single_shot;

private:
    // Our sleep control variables.
    std::mutex timerMutex;
    std::condition_variable timerCond;
    bool stopThread;
};

} /* namespace FAS_THREAD */

#endif /* FI_TICKER_H_ */
