/*!
 * @file fi_tiff.h
 * @brief This file contains the definitions and declarations for
 *        Fastec [TIFF](\ref GLOSSARY_TIFF) images.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 *
 * @sa https://www.awaresystems.be/imaging/tiff/faq.html
 *
 * @b Abbreviations
 * @{
 * @anchor GLOSSARY_IFD
 * @li IFD: Image File Directory.
 * @anchor GLOSSARY_TIFF
 * @li TIFF: Tagged Image File Format.
 * @}
 *
 */

#ifndef FIVID_FI_TIFF_
#define FIVID_FI_TIFF_
//#pragma once

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <memory>
#include <vector>

#include "fas_video.h"
#include "fi_image.h"

namespace FI_TIFF
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

//! Align the [IFD](\ref GLOSSARY_IFD) and tag data on a 4 byte boundary so that we don't have any issues
//! with structure alignment.
const uint32_t TIFF_IFD_DATA_ALIGNMENT  = 4;

//! TIFF magic constant.
const uint16_t TIFF_MAGIC               = 42;

//! @name Tag Types.
//!
//! These are the constants used to define what "type" of data a tag contains.
//! They aren't in an "enum" because they need to have a specific value.
//! @{
const uint16_t TIFF_TYPE_BYTE          = 1;
const uint16_t TIFF_TYPE_ASCII         = 2;
const uint16_t TIFF_TYPE_SHORT         = 3;
const uint16_t TIFF_TYPE_LONG          = 4;
const uint16_t TIFF_TYPE_RATIONAL      = 5;
const uint16_t TIFF_TYPE_SBYTE         = 6;
const uint16_t TIFF_TYPE_UNDEFINED     = 7;
const uint16_t TIFF_TYPE_SSHORT        = 8;
const uint16_t TIFF_TYPE_SLONG         = 9;
const uint16_t TIFF_TYPE_SRATIONAL     = 10;
const uint16_t TIFF_TYPE_FLOAT         = 11;
const uint16_t TIFF_TYPE_DOUBLE        = 12;
//! @}

//! @name Tag Types Sizes.
//!
//! Definitions for the lengths of the above values.  Note that "ASCII" and UNDEFINED
//! are not listed.
//! @{
const uint32_t TIFF_BYTE_SIZE          = (sizeof(uint8_t));
const uint32_t TIFF_SHORT_SIZE         = (sizeof(uint16_t));
const uint32_t TIFF_LONG_SIZE          = (sizeof(uint32_t));
const uint32_t TIFF_RATIONAL_SIZE      = (2 * TIFF_LONG_SIZE);
const uint32_t TIFF_SBYTE_SIZE         = (sizeof(int8_t));
const uint32_t TIFF_SSHORT_SIZE        = (sizeof(int16_t));
const uint32_t TIFF_SLONG_SIZE         = (sizeof(int32_t));
const uint32_t TIFF_SRATIONAL_SIZE     = (2 * TIFF_SLONG_SIZE);
//! @}

//! @name Tag Function Macros
//!
//! These are macros used to generate the function names for the
//! tag size and tag read functions.
//! @{
#define TIFF_TAG_SIZE_FUNCTION_NAME(x)  tag_size##x
#define TIFF_TAG_READ_FUNCTION_NAME(x)  tag_read##x
//! @}

//! @name Miscellaneous Macros
//! @{

//! Use this macro to fill in an array of tag entry elements.
//! 'x' is the tag number.
//! 't' is the type of the tag value for that number.
#define TAG_DB_ENTRY(x,t) tagEntry(this, x, t, TIFF_TAG_SIZE_FUNCTION_NAME(x), TIFF_TAG_READ_FUNCTION_NAME(x))

//! The "SIZE" functions return the size of the tag data increased to the next
//! multiple of FI_TIFF::TIFF_IFD_DATA_ALIGNMENT.  If the tag is not to be used for a given
//! format, a size of 0 should be returned.
#define TIFF_TAG_SIZE_FUNCTION(x) static uint64_t TIFF_TAG_SIZE_FUNCTION_NAME(x)(void *tiff, uint32_t fourcc, const pm_metadata *meta)

//! The "READ" functions read the tag data into the buffer and fill in the
//! ifd_ptr->count field.
#define TIFF_TAG_READ_FUNCTION(x) static void TIFF_TAG_READ_FUNCTION_NAME(x)(void *tiff, uint32_t fourcc, const pm_metadata *meta, IFD_ENTRY *ifd_ptr, char *buffer)

//! Aligns a size up to the next multiple of FI_TIFF::TIFF_IFD_DATA_ALIGNMENT.
#define TIFF_ALIGNED_SIZE(s)    (((s) + (FI_TIFF::TIFF_IFD_DATA_ALIGNMENT - 1)) & (~(FI_TIFF::TIFF_IFD_DATA_ALIGNMENT - 1)))
//! @}

#pragma pack(1)
//! The first part of a TIFF image.
typedef struct
{
    uint16_t byte_order;
    uint16_t tif_magic;
    uint32_t ifd_offset;
} TIFF_HEADER;
#pragma pack(0)

#pragma pack(1)
//! An IFD is composed of a 16 bit count followed by the IFD_ENTRY array, followed
//! by a 32 bit count which is the offset of the next IFD or 0 if there is none.
typedef struct
{
    uint16_t tag;
    uint16_t type;
    uint32_t count;
    uint32_t val_offset;
} IFD_ENTRY;
#pragma pack(0)

//! Prototype of the tag functions returning the number of bytes required for the tag entry.
typedef uint64_t (* TAG_SIZE_FN)(void *tiff, uint32_t fourcc, const pm_metadata *meta);

//! Prototype of the tag functions reading the tag data into the location pointed to by 'buffer'.
typedef void (* TAG_READ_FN)(void *tiff, uint32_t fourcc, const pm_metadata *meta, IFD_ENTRY *ifd_ptr, char *buffer);

//--------------------------------------------------------------------
// Forward Declaration
//--------------------------------------------------------------------

class FiTIFF;

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

/*!
 * @brief Table of routines to fill in and get size for TIFF tags.
 *
 * When generating a TIFF file, a vector of TAG_ENTRY structures is iterated
 * to read the tag data and query the size of the file.
 *
 * The total size of a TIFF file will be the image size + TIFF header + ifd +
 * the sum of all tags with a size greater than 4 (since values that take up
 * 4 bytes or less are stored directly in the IFD entry.
 */
class tagEntry
{
public:
    explicit tagEntry(void *parent, uint16_t tag, uint16_t type, TAG_SIZE_FN size_fn, TAG_READ_FN read_fn);
    ~tagEntry() {}

    uint16_t get_id() const     { return tag_id; }
    uint16_t get_type() const   { return tag_type; }

    uint64_t size(uint32_t fourcc, const pm_metadata *meta);
    void     read(uint32_t fourcc, const pm_metadata *meta, IFD_ENTRY *ifd_ptr, char *buffer);

    bool operator<(const tagEntry& t1) const { return (tag_id < t1.tag_id); }
    bool operator<=(const tagEntry& t1) const { return (tag_id <= t1.tag_id); }
    bool operator>(const tagEntry& t1) const { return (tag_id > t1.tag_id); }
    bool operator>=(const tagEntry& t1) const { return (tag_id >= t1.tag_id); }
    bool operator==(const tagEntry& t1) const { return (tag_id == t1.tag_id); }
    bool operator!=(const tagEntry& t1) const { return (tag_id != t1.tag_id); }

private:
    FiTIFF *tiff;
    uint16_t tag_id;
    uint16_t tag_type;
    TAG_SIZE_FN tag_size;
    TAG_READ_FN tag_read;
};


// The main TIFF class.
class FiTIFF : public FAS_VIDEO::FiImage
{
public:
    FiTIFF();
    virtual ~FiTIFF() {}

    virtual int CreateHeader(uint32_t fourcc, const pm_metadata *meta) override;
    virtual std::unique_ptr<FAS_VIDEO::HEADER_DATA> GetHeader() override;

    // The public property accessor functions.
    virtual FAS_VIDEO::IMAGE_FORMAT GetImageFormat() override;

    virtual uint32_t get_header_size(uint32_t fourcc, const pm_metadata *meta);
    virtual uint32_t get_ifd_data_size(uint32_t fourcc, const pm_metadata *meta);
    virtual uint32_t get_image_size(uint32_t fourcc, const pm_metadata *meta);
    virtual uint16_t get_photometric_interpretation(uint32_t fourcc);

protected:
    virtual bool add_tags(uint32_t fourcc, const pm_metadata *meta, uint32_t ifd_offset, uint32_t data_offset);
    virtual void init_tag_db();
    virtual bool VerifyImageData(const uint8_t *imageDataToVerify, uint32_t imageDataSize) override;

protected:
    std::vector<tagEntry> tag_db;   //!< Database of tags for this image.
    std::vector<uint8_t> header;    //!< Array of bytes for the "total" header.
};

} // End of FI_TIFF

#endif  /* FIVID_FI_TIFF_ */

