/*!
 * @file lince5m_param.h
 * @brief This is the header file providing the hardware abstraction layer
 * for the Lince5M and its sensor-specific FPGA interface.
 *
 * Copyright 2010, 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */
#ifndef __LINCE5M_PARAM_H__
#define __LINCE5M_PARAM_H__
#ifdef __cplusplus
extern "C" {
#endif

#define PIXELS_PER_PPW 16   //!< Sensor groups pixels into PPWs "pixel processing windows".
#define PPWS_PER_BLOCK 4    //!< Sensor emits multiples of 64 pixels, which is 4 PPWs.

//! First line of defect map file.
#define LINCE5M_DEFECT_HEADER "AL5Xv1:%u\n"

//! Maximum number of pixel defects allowed. Anafocus manufacturing permits up
//! to 100 whitespot pixels and up to 150 black/bright pixels.
#define MAX_DEFECTS 250

//! The sensor emits pixels in groups of PPWs and wants the number of PPWs to be
//! multiple of 4, so we tell the FPGA how to extract pixels from these blocks of 64.
#define PIXELS_PER_BLOCK  ( PIXELS_PER_PPW * PPWS_PER_BLOCK )

//! Lince5M sensor actual size.
#define LINCE_SENSOR_WIDTH	2560
#define LINCE_SENSOR_HEIGHT	2048

//! The maximum size image we allow, what we expose to the user.
#define LINCE_REPORTED_MAX_WIDTH  LINCE_SENSOR_WIDTH
#define LINCE_REPORTED_MAX_HEIGHT LINCE_SENSOR_HEIGHT

// Maximum ROI we permit in 10-bit mode
#define LINCE_MAX_WIDTH_10BIT   LINCE_SENSOR_WIDTH
#define LINCE_MAX_HEIGHT_10BIT  LINCE_SENSOR_HEIGHT

// Maximum ROI we permit in 12-bit mode
#define LINCE_MAX_WIDTH_12BIT   LINCE_SENSOR_WIDTH
#define LINCE_MAX_HEIGHT_12BIT  LINCE_SENSOR_HEIGHT

//! Minimum ROI we are supporting currently.
#ifdef SAUTIL
#define LINCE_ROI_MIN_WIDTH	64
#define LINCE_ROI_MIN_HEIGHT	 4
#else
#define LINCE_ROI_MIN_WIDTH	64
#define LINCE_ROI_MIN_HEIGHT	32
#endif

//! Default frame rate when none has been set.
#define LINCE_DEF_FRAME_RATE	250

//! Symbols for t_row values
#define TROW_52   52
#define TROW_84   84
#define TROW_124 124
#define TROW_132 132
#define TROW_165 165

//! Maximum width to use register files with faster row timing.
#define TROW_52_MAX_WIDTH   768
#define TROW_84_MAX_WIDTH  1280
#define TROW_124_MAX_WIDTH 1920
#define TROW_132_MAX_WIDTH 2048

//! Values from Lince5M_Regs_FM1_86.4MHz_trow{52,84}_HSHS[_bin2x2]_rev910.dat
//! The values for [_bin2x2] for trow=84 are the same as the normal mode file.
#define FM1_N_CYCLES_B_52   0x000000cb	//! Rg0x004-Rg0x005 for t_row=52. (203)
#define FM1_N_CYCLES_F_52   0x000002a8	//! Rg0x006-Rg0x007 for t_row=52. (680)
#define FM1_N_CYCLES_B_84   0x00000075	//! Rg0x004-Rg0x005 for t_row=84. (117)
#define FM1_N_CYCLES_F_84   0x000000fb	//! Rg0x006-Rg0x007 for t_row=84. (251)
//! Values from Lince5M_Regs_FM1_86.4MHz_trow{124,132,165}_HSHS_rev707.dat
#define FM1_N_CYCLES_B_124  0x000000cb	//! Rg0x004-Rg0x005 for t_row=124. (203)
#define FM1_N_CYCLES_F_124  0x000003fe	//! Rg0x006-Rg0x007 for t_row=124. (1022)
#define FM1_N_CYCLES_B_132  0x000000cb	//! Rg0x004-Rg0x005 for t_row=132. (203)
#define FM1_N_CYCLES_F_132  0x000003fe	//! Rg0x006-Rg0x007 for t_row=132. (1022)
#define FM1_N_CYCLES_B_165  0x000000cb	//! Rg0x004-Rg0x005 for t_row=165. (203)
#define FM1_N_CYCLES_F_165  0x0000035a	//! Rg0x006-Rg0x007 for t_row=165. (858)
//! Values from Lince5M_Regs_FM6_72MHz_trow84_HSHS_rev914.dat
//! The values for [_bin2x2] for trow=84 are the same as the normal mode file.
#define FM6_N_CYCLES_B_84   0x00000062	//! Rg0x004-Rg0x005 for t_row=84. (98)
#define FM6_N_CYCLES_F_84   0x000000d3	//! Rg0x006-Rg0x007 for t_row=84. (211)
//! Values from Lince5M_Regs_FM6_72MHz_trow{124,132,165}_HSHS_rev{559,558,622}.dat
#define FM6_N_CYCLES_B_124  0x000000a8	//! Rg0x004-Rg0x005 for t_row=124. (168)
#define FM6_N_CYCLES_F_124  0x000002df	//! Rg0x006-Rg0x007 for t_row=124. (735)
#define FM6_N_CYCLES_B_132  0x000000a8	//! Rg0x004-Rg0x005 for t_row=132. (168)
#define FM6_N_CYCLES_F_132  0x000002df	//! Rg0x006-Rg0x007 for t_row=132. (735)
#define FM6_N_CYCLES_B_165  0x000000a8	//! Rg0x004-Rg0x005 for t_row=165. (168)
#define FM6_N_CYCLES_F_165  0x000002d3	//! Rg0x006-Rg0x007 for t_row=165. (723)

#define EXTRA_ROWS   2	//! Rg0x00A, the same for all files in both framing modes.
#define EXTRA_CYCLES 1	//! Rg0x00b-Rg0x00c, the same for all files in both framing modes.

// Even if these are not needed for the common utilities, include them here just
// so all the values we have to extract from the register files are together.

// Values from Lince5M_Regs_FM1_86.4MHz_trow{52,84}_HSHS[_bin2x2]_rev910.dat
// The values for [_bin2x2] for trow=84 are the same as the normal mode file.
#define FM1_ANALOG_GAIN_52  0x001E  //! Rg0x05B for t_row=52.
#define FM1_ANALOG_GAIN_84  0x001E  //! Rg0x05B for t_row=84.
// Values from Lince5M_Regs_FM1_86.4MHz_trow{124,132,165}_HSHS_rev707.dat
#define FM1_ANALOG_GAIN_124 0x0020  //! Rg0x05B for t_row=124.
#define FM1_ANALOG_GAIN_132 0x0020  //! Rg0x05B for t_row=132.
#define FM1_ANALOG_GAIN_165 0x001E  //! Rg0x05B for t_row=165.
// Values from Lince5M_Regs_FM6_72MHz_trow84_HSHS[_bin2x2]_rev910.dat
// The values for [_bin2x2] for trow=84 are the same as the normal mode file.
#define FM6_ANALOG_GAIN_84  0x001E  //! Rg0x05B for t_row=84.
// Values from Lince5M_Regs_FM6_72MHz_trow{124,132,165}_HSHS_rev{559,558,622}.dat
#define FM6_ANALOG_GAIN_124 0x0020  //! Rg0x05B for t_row=124.
#define FM6_ANALOG_GAIN_132 0x0020  //! Rg0x05B for t_row=132.
#define FM6_ANALOG_GAIN_165 0x001D  //! Rg0x05B for t_row=165.

//! Lince Frame Overhead Time is the duration of State F.
//! For FM1, max value of N_CYCLES_F in Rg0x006-Rg0x007 is 1022. At 86.4MHz, that's 11.8 usecs.
//! For FM6, max value of N_CYCLES_F in Rg0x006-Rg0x007 is 735. At 72MHz, that's 10.2 usecs.
//! We take the worst case and use 12 usecs. Using too large a value means that maybe
//! we don't eke out one more frame per second in some case.
#define LINCE_FOT_USEC	    12

//! Lince minimum exposure time is the duration of State B which is no exposure at all.
//! For FM1, max value of N_CYCLES_B in Rg0x004-Rg0x005 is 203. At 86.4MHz, that's 2.35 usecs.
//! For FM6, max value of N_CYCLES_B in Rg0x004-Rg0x005 is 168. At 72MHz, that's 2.33 usecs.
//! But for trow=52 where we don't have separate files for FM1 vs FM6, N_CYCLES_B is 203.
//! At 72Mhz, that's 2.82 usecs -- lucky break, still under 3 usecs!
//! We can just use the same value for all framing modes.
#define LINCE_MIN_EXP_USEC  3

//! f_sys is the system frequency, which depends on the framing mode.
#define FM1_F_SYS 86400000  // 86.4 MHz for Framing Mode 1
#define FM6_F_SYS 72000000  // 72.0 MHz for Framing Mode 6

/*
 * Rg0x000 = trigger configuration
 * Rg0x001 = device status
 * Rg0x002-Rg0x003 = integration time
 * Rg0x004-Rg0x005 = N.cycles_B, duration of State B.
 * Rg0x006-Rg0x007 = N.cycles_F, duration of State F.
 * Rg0x00a = extra rows to adjust frame rate by adding vertical blanking
 * Rg0x00b-Rg0x00c = extra cycles to add to readout to adjust frame rate
 * Rg0x010-0x013 = ROI (row, column, num rows, num cols)
 * Rg0x014 = subsampling and binning
 * Rg0x015 = t.row_min_cycles, t.row_cycles, can increase to slow frame rate
 * Rg0x017-Rg0x20 = Global digital gains and offsets
 * Rg0x01c = used in VFPN calibration
 * Rg0x024 = LVDS interface configuration
 * Rg0x025-Rg0x026, Rg0x036, Rg0x038 - bits per pixel and LVDS pad configuration
 * Rg0x027-Rg0x033 = LVDS port skews
 * Rg0x034 = IDLE pattern
 * Rg0x036 = coarse delays for LVDS ports
 * Rg0x038,Rg0x039 = framing mode configuration and IDLE pattern for CMOS
 * Rg0x048-Rg0x049 = used for calibration in "strong light conditions" and for HDR
 * Rg0x04a = used for HDR configuration
 * Rg0x04b = either 0x1740 or 0x174c depending on whether sensor is covered in calibration
 * Rg0x04c = used for HDR configuration
 * Rg0x04d = used for calibration in HDR mode
 * Rg0x05b = analog gain
 * Rg0x065 = analog offset
 * Rg0x06a = used for HDR configuration
 * Rg0x06b = used for HDR configuration
 * Rg0x070 = used for HDR configuration
 * Rg0x07c = polarization current of the pixel, increased for binning with stressed trow
 * Rg0x085 = used for HDR configuration
 * Rg0x087-Rg0x091 = modify from defaults for HDR configuration (3.4.2.1)
 * Rg0x0e8-Rg0x0e9 = used for calibration in "strong light conditions"
 * Rg0x14d-Rg0x14e = for adding horizontal blanking
 * Rg0x151-Rg0x152 = component of fixed latency between trigger start and exposure start
 * Rg0x153-Rg0x154 = component of fixed latency between trigger start and exposure start
 * Rg0x155-Rg0x156 = component of fixed latency between trigger start and exposure start
 * Rg0x157-Rg0x158 = component of minimum time trigger must remain inactive
 * Rg0x159-Rg0x15a = component of minimum time trigger must remain inactive
 * Rg0x15b-Rg0x15c = component of minimum time trigger must remain inactive
 * Rg0x15d-Rg0x15e = component of minimum time trigger must remain inactive
 * Rg0x15f-Rg0x160 = component of minimum time trigger must remain inactive
 * Rg0x168 = modify from defaults for HDR configuration (3.4.2.1)
 * Rg0x170-Rg0x171 = control programmable memories access
 * Rg0x172 = constant latency, minimum cycles between trigger starts + 1 (single-edge mode)
 * Rg0x174 = used for calibration in HDR mode
 * Rg0x176, Rg0x180 = to support Hot ROI
 * Rg0x17a = write 0x3e to enable readout of top dummy rows
 * Rg0x17e-Rg0x17f = disable groups of optically black pixels, e.g. if defective
 * Rg0x180 = cfg_all_array, force enable all active array even if outside ROI
 * Rg0x1a9 = compensation digital offset
 * Rg0x1aa, Rg0x1ab = something about VFPN offset
 * Rg0x1ad = unlock VFPN offset correction memories (0x0a00-0x13ff)
 * Rg0x1ae = control defective pixel correction
 * Rg0x1b5 = has enable for HFPN offset correction
 * Rg0x1b7 = enable global digital gains and offsets
 * Rg0x1bc = enable statistics module of DIP
 * Rg0x1cb-Rg0x1cc = read OB column sum
 */

//! Convert integration in microseconds to sys clock counts
//! In Framing Mode 1, f.sys = 86.4 MHz.
#define FM1_INTEGRATION_USECS_TO_FSYS( u ) (( (u) * 864 + 5 ) / 10 )
//! In Framing Mode 6, f.sys = 72.0 MHz.
#define FM6_INTEGRATION_USECS_TO_FSYS( u ) (( (u) * 720 + 5 ) / 10 )

#ifdef __cplusplus
}
#endif
#endif /*__LINCE5M_PARAM_H__ */

