/*!
 * @file fpga_io.c
 * @brief Register I/O operations for the FPGA device driver.
 *
 * Copyright (C) 2010,2015 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <linux/device.h>
#include <linux/io.h>
#include <linux/clk.h>
#include <linux/delay.h>

#include "fi_pci.h"
#include "fi_fpga_dev.h"

/*!
 * @brief Write a value to a FPGA register.
 *
 * @param dev Pointer to our device structure.
 * @param reg_id The register we want to write to.
 * @param reg_value The value we want to write to the register.
 */
void fi_set_fpga_reg(FI_FPGA_DEV *dev, uint32_t reg_id, uint32_t reg_value)
{
    ulong   flags;
    uint32_t    *ioaddr;

    // Sanity check for a null device pointer.
    if (dev == NULL)
        return;

    // Sanity check register ID value.
    if ((reg_id < FI_FPGA_BASE_ADDR) || (reg_id > FI_FPGA_MAX_ADDR))
    {
        /* If the caller dereferences the zero we return, we'll get a kernel oops. */
        dev_err(dev->fpga_dev, "%s: Bad reg_id 0x%08x!\n", __FUNCTION__, reg_id);
        return;
    }

    // Write the value to the register.
    ioaddr = fi_pci_get_base();
    if (ioaddr == NULL)
    {
        dev_err(dev->fpga_dev, "%s: No base address available!\n", __FUNCTION__);
        return;
    }

    // The register ID is a byte offset value. We're adding it
    // to a 32-bit pointer, so we need to make the ID a word offset.
    reg_id >>= 2;

    // Perform the write.
    spin_lock_irqsave(&dev->reg_lock, flags);
    iowrite32(reg_value, ioaddr + reg_id);
    spin_unlock_irqrestore(&dev->reg_lock, flags);
}

/*!
 * @brief Read a value from a FPGA register.
 *
 * @param dev Pointer to our device structure.
 * @param reg_id The register we want to read from.
 *
 * @return The value read from reg_id.
 */
uint32_t fi_get_fpga_reg(FI_FPGA_DEV *dev, uint32_t reg_id)
{
    uint32_t    ret_val;
    ulong   flags;
    uint32_t    *ioaddr;

    // Sanity check for a null device pointer.
    if (dev == NULL)
        return reg_id;

    // Sanity check register ID value.
    if ((reg_id < FI_FPGA_BASE_ADDR) || (reg_id > FI_FPGA_MAX_ADDR))
    {
        /* If the caller dereferences the zero we return, we'll get a kernel oops. */
        dev_err(dev->fpga_dev, "%s: Bad reg_id 0x%08x!\n", __FUNCTION__, reg_id);
        return 0;
    }

    // Read the value from the register.
    ioaddr = fi_pci_get_base();
    if (ioaddr == NULL)
    {
        dev_err(dev->fpga_dev, "%s: No base address available!\n", __FUNCTION__);
        return 0;
    }

    // The register ID is a byte offset value. We're adding it
    // to a 32-bit pointer, so we need to make the ID a word offset.
    reg_id >>= 2;

    spin_lock_irqsave(&dev->reg_lock, flags);
    ret_val = ioread32(ioaddr + reg_id);
    spin_unlock_irqrestore(&dev->reg_lock, flags);

    return ret_val;
}

/*!
 * @brief Peform a Read-Modify-Write operation on the FPGA.
 *
 * reg_value = ( reg_value & mask ) | value;
 *
 * @param reg_addr  Our FPGA register we'll be modifying.
 * @param value     Value to set into the register.
 * @param mask      Value to mask with the existing contents of the register.
 *
 * Performs
 *
 *  new_reg = ( old_reg & mask ) | value
 *
 * To set a bit
 *
 *  fpga_modify_reg( reg_name, BIT_NAME, -1 );
 *
 * To clear a bit
 *
 *  fpga_modify_reg( reg_name, 0, ~BIT_NAME );
 *
 * Example of changing a set of bits:
 *
 *  value =    ILC_BIT_DEPTH_8    | ILC_BIT_SEL_11_4;
 *  mask  = ~( ILC_BIT_DEPTH_MASK | ILC_BIT_SEL_MASK );
 *  fpga_modify_reg( reg_name, value, mask );
 *
 * @return true if everything went well, otherwise false.
 */
uint32_t fi_mod_fpga_reg(FI_FPGA_DEV *dev, uint32_t reg_id, uint32_t reg_value, uint32_t reg_mask)
{
    uint32_t    temp_value;

    // Sanity check for a null device pointer.
    if (dev == NULL)
        return ((uint32_t) - EFAULT);

    // Make sure the register ID makes sense.
    if (!FI_FPGA_REG_VALID(reg_id))
    {
        dev_err(dev->fpga_dev, "Bad modify reg 0x%08x\n", reg_id);
        return ((uint32_t) - EFAULT);
    }

    // Grab the info from the FPGA.
    temp_value = fi_get_fpga_reg(dev, reg_id);

    // Mask, modify and write it back out.
    reg_value = (temp_value & reg_mask) | reg_value;

    // Update the register on the FPGA.
    fi_set_fpga_reg(dev, reg_id, reg_value);

    // All is good.
    return reg_value;
}

/*!
 * @brief Wait for the I2C bus to go IDLE.
 *
 * @param data_ptr [optional] Pointer to where to put the read data.
 * @param bus      Which bus to talk on.
 * @return True if success, false otherwise.
 */
static bool i2c_wait_idle(FI_FPGA_DEV *dev, uint8_t *data_ptr, int bus)
{
    bool rc;
    int retry_count;
    uint32_t status_reg;
    uint32_t i2c_status = I2CM_STATUS_BUSY;
    const int max_retries = 10;

//    printk("%s\n", __FUNCTION__);

    // Which bus are we talking to?
    if (bus == FI_I2C_MAIN)
        status_reg = I2C_MASTER_STATUS;
    else if (bus == FI_I2C_IO)
        status_reg = I2C_MASTER_STATUS_IOB;
    else
        status_reg = I2C_MASTER_STATUS_ISB;


    // Control access to the driver.
    for (retry_count = 0; retry_count < max_retries; ++retry_count)
    {
        // Check to see if the busy bit is still set.
        i2c_status = fi_get_fpga_reg(dev, status_reg);
        if ((i2c_status & I2CM_STATUS_BUSY) == 0)
            break;

        // Wait for a little bit.
        usleep_range(500, 2000);
    }

    if (i2c_status & I2CM_STATUS_BUSY)
    {
        printk("%s: timeout: 0x%08x\n", __FUNCTION__, i2c_status);
    }
    else if ((i2c_status & I2CM_STATUS_ERR) != 0)
    {
        printk("%s: error bit set: 0x%08x\n", __FUNCTION__, i2c_status);
    }
    else if (data_ptr != NULL)
    {
        uint32_t control_val;
        uint32_t control_reg;
        uint32_t data_reg;
        uint32_t data_val;

        // Which bus are we talking to?
        if (bus == FI_I2C_MAIN)
        {
            control_reg = I2C_MASTER_CONTROL;
            data_reg = I2C_MASTER_RD_DATA;
        }
        else if (bus == FI_I2C_IO)
        {
            control_reg = I2C_MASTER_CONTROL_IOB;
            data_reg = I2C_MASTER_RD_DATA_IOB;
        }
        else
        {
            control_reg = I2C_MASTER_CONTROL_ISB;
            data_reg = I2C_MASTER_RD_DATA_ISB;
        }

        // Read the data registers.
        data_val = fi_get_fpga_reg(dev, data_reg);

        // Do we have 1 or 2 bytes?
        control_val = fi_get_fpga_reg(dev, control_reg);
        if ((control_val & I2CM_CMD_4DATA_CNT) == I2CM_CMD_4DATA_CNT)
        {
            *data_ptr = ((data_val >> 24) & 0xFF);
            ++data_ptr;
            *data_ptr = ((data_val >> 16) & 0xFF);
            ++data_ptr;
            *data_ptr = ((data_val >> 8) & 0xFF);
            ++data_ptr;
        }
        else if ((control_val & I2CM_CMD_2DATA_CNT) == I2CM_CMD_2DATA_CNT)
        {
            *data_ptr = ((data_val >> 8) & 0xFF);
            ++data_ptr;
        }

        *data_ptr = (data_val & 0xFF);
    }

    // Did we exceed our retries?
    if (data_ptr == NULL)
        rc = ((i2c_status & I2CM_STATUS_BUSY) == 0);
    else
        rc = ((i2c_status & (I2CM_STATUS_BUSY | I2CM_STATUS_ERR)) == 0);

    // Return our result.
    return rc;
}

/*!
 * @brief Logic to perform an I2C read of a register from a device.
 *
 * @param dev        Pointer to the device driver structure.
 * @param chip_addr  Address of the device we're interested in.
 * @param reg_addr   Address of the register we want to write, or other commands to be sent.
 * @param reg_data   Pointer to where to put the data.
 * @param bus        Which bus to talk on.
 * @param byte_count    Number of bytes we want to read.
 * @param cmd_count     Number of bytes in the command (most read operations require a command to be sent first).
 * @param readonly      True if we are only performing a read operations (most reads require you to write the register address).
 *
 * @return 0 if success, error code otherwise.
 */
uint32_t fi_i2c_read(FI_FPGA_DEV *dev, uint8_t chip_addr, uint32_t reg_addr, uint8_t *reg_data, int bus, int byte_count, int cmd_count, bool readonly)
{
    uint32_t    rc;
    uint32_t    reg_value;
    uint32_t    control_reg;
    uint32_t    cmd_data_reg;

//    printk("%s: chip=0x%x, reg=0x%x, bus=%d, byte_count=%d, cmd_count=%d, ro=%d\n",
//      __FUNCTION__, chip_addr, reg_addr, bus, byte_count, cmd_count, readonly);

    // Make sure the I2C bus is available.
    if (!i2c_wait_idle(dev, NULL, bus))
    {
        printk(KERN_ERR "%s: I2C never becomes idle.\n", __FUNCTION__);
        rc = ((uint32_t) - EFAULT);
        goto bail;
    }

    // Which bus are we talking to?
    if (bus == FI_I2C_MAIN)
    {
        control_reg = I2C_MASTER_CONTROL;
        cmd_data_reg = I2C_MASTER_CMD_DATA;
    }
    else if (bus == FI_I2C_IO)
    {
        control_reg = I2C_MASTER_CONTROL_IOB;
        cmd_data_reg = I2C_MASTER_CMD_DATA_IOB;
    }
    else
    {
        control_reg = I2C_MASTER_CONTROL_ISB;
        cmd_data_reg = I2C_MASTER_CMD_DATA_ISB;
    }

    // Read the data.
    if (readonly)
    {
        // Read-only command to the FPGA.
        reg_value = I2CM_CMD_READ_ONLY;
    }
    else
    {
        // The command is write+read.
        reg_value = I2CM_CMD_READ;
    }

    // Do we have a command to send?
    if (cmd_count > 0)
    {
        // Write out the command bytes (typically the register address we're interested in).
        fi_set_fpga_reg(dev, cmd_data_reg, reg_addr);
        if (cmd_count == 4)
            reg_value |= I2CM_CMD_4CMD_CNT;
        else
            reg_value |= I2CM_CMD_1CMD_CNT;
    }

    reg_value |= I2CM_CMD_DEV(chip_addr);
    if (byte_count == 4)
        reg_value |= I2CM_CMD_4DATA_CNT;
    else if (byte_count == 2)
        reg_value |= I2CM_CMD_2DATA_CNT;
    else if (byte_count == 1)
        reg_value |= I2CM_CMD_1DATA_CNT;
    fi_set_fpga_reg(dev, control_reg, reg_value);

    // Now wait for the data to come back.
    if (!i2c_wait_idle(dev, reg_data, bus))
    {
        printk(KERN_ERR "%s: I2C never becomes idle or there was an error.\n", __FUNCTION__);
        rc = ((uint32_t) - EFAULT);
        goto bail;
    }

#if 0
    if (byte_count == 4)
        printk("[%02X] = 0x%02X%02X%02X%02X\n", reg_addr, reg_data[3], reg_data[2], reg_data[1], reg_data[0]);
    else if (byte_count == 2)
        printk("[%02X] = 0x%02X%02X\n", reg_addr, reg_data[1], reg_data[0]);
    else
        printk("[%02X] = 0x%02X\n", reg_addr, reg_data[0]);
#endif

    // Life is good.
    rc = 0;

bail:

    // Return our status.
    return rc;
}

/*!
 * @brief Logic to perform an I2C write to a register of a device.
 *
 * @param chip_addr Address of the device we're interested in.
 * @param cmd_data Address of the register we want to write, or other commands to be sent.
 * @param reg_data Data we want written.
 *
 * @return 0 if success, error code otherwise.
 */
uint32_t fi_i2c_write(FI_FPGA_DEV *dev, uint8_t chip_addr, uint32_t cmd_data, uint32_t reg_data, int bus, int byte_count, int cmd_count)
{
    uint32_t rc;
    uint32_t reg_value;

//    printk("%s: chip_addr=0x%02x, cmd_data=0x%08x/%d, reg_data=0x%08x/%d, bus=%d", __FUNCTION__, chip_addr, cmd_data, cmd_count, reg_data, byte_count, bus);

    // Make sure the I2C bus is available.
    if (!i2c_wait_idle(dev, NULL, bus))
    {
        printk(KERN_ERR "%s: I2C never becomes idle.\n", __FUNCTION__);
        rc = ((uint32_t) - EFAULT);
        goto bail;
    }

    // Which bus are we talking to?
    if (bus == FI_I2C_MAIN)
    {
        // Write the data, if any.
        if (byte_count > 0)
        {
            fi_set_fpga_reg(dev, I2C_MASTER_WR_DATA, reg_data);
        }

        // Write the command.
        fi_set_fpga_reg(dev, I2C_MASTER_CMD_DATA, cmd_data);

        // Build up the command register.
        reg_value = I2CM_CMD_WRITE;
        reg_value |= I2CM_CMD_DEV(chip_addr);
        if (cmd_count == 4)
            reg_value |= I2CM_CMD_4CMD_CNT;
        else
            reg_value |= I2CM_CMD_1CMD_CNT;
        if (byte_count == 4)
            reg_value |= I2CM_CMD_4DATA_CNT;
        else if (byte_count == 2)
            reg_value |= I2CM_CMD_2DATA_CNT;
        else if (byte_count == 1)
            reg_value |= I2CM_CMD_1DATA_CNT;

        fi_set_fpga_reg(dev, I2C_MASTER_CONTROL, reg_value);
    }
    else if (bus == FI_I2C_IO)
    {
        // Write the data, if any.
        if (byte_count > 0)
        {
            fi_set_fpga_reg(dev, I2C_MASTER_WR_DATA_IOB, reg_data);
        }

        // Write the command.
        fi_set_fpga_reg(dev, I2C_MASTER_CMD_DATA_IOB, cmd_data);

        // Build up the command register.
        reg_value = I2CM_CMD_WRITE;
        reg_value |= I2CM_CMD_DEV(chip_addr);
        if (cmd_count == 4)
            reg_value |= I2CM_CMD_4CMD_CNT;
        else
            reg_value |= I2CM_CMD_1CMD_CNT;
        if (byte_count == 4)
            reg_value |= I2CM_CMD_4DATA_CNT;
        else if (byte_count == 2)
            reg_value |= I2CM_CMD_2DATA_CNT;
        else if (byte_count == 1)
            reg_value |= I2CM_CMD_1DATA_CNT;

        fi_set_fpga_reg(dev, I2C_MASTER_CONTROL_IOB, reg_value);
    }
    else
    {
        // Write the data, if any.
        if (byte_count > 0)
        {
            fi_set_fpga_reg(dev, I2C_MASTER_WR_DATA_ISB, reg_data);
        }

        // Write the command.
        fi_set_fpga_reg(dev, I2C_MASTER_CMD_DATA_ISB, cmd_data);

        // Build up the command register.
        reg_value = I2CM_CMD_WRITE;
        reg_value |= I2CM_CMD_DEV(chip_addr);
        if (cmd_count == 4)
            reg_value |= I2CM_CMD_4CMD_CNT;
        else
            reg_value |= I2CM_CMD_1CMD_CNT;
        if (byte_count == 4)
            reg_value |= I2CM_CMD_4DATA_CNT;
        else if (byte_count == 2)
            reg_value |= I2CM_CMD_2DATA_CNT;
        else if (byte_count == 1)
            reg_value |= I2CM_CMD_1DATA_CNT;

        fi_set_fpga_reg(dev, I2C_MASTER_CONTROL_ISB, reg_value);
    }

    // Life is good.
    rc = 0;

bail:
    // Return our result.
    return rc;
}


