/*!
 * @file fi_pci_fops.c
 * @brief File operation functions for the Fastec PCI device driver.
 *
 * Copyright (C) 2016 Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/types.h>    /* size_t */
#include <linux/errno.h>    /* error codes */
#include <linux/fs.h>       /* everything... */
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/pfn_t.h>
#include <linux/printk.h>
#include <linux/uaccess.h>

#include "fi_pci_dev.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

//! The maximum number of minor device nodes we'll support.
#define MAX_MINOR       1

//! Our virtual memory management structure.
typedef struct FI_PCI_VMA
{
    FASTEC_PCI_PRIVATE  *fi_dev;	//!< Pointer to our driver structure.
    atomic_t mmap_count;		//!< Number of mmap's currently in operation.
    size_t nr_pages;			//!< The number of pages we have mapped.

    struct mutex page_mutex;		//!< Access control mutex for the used_pages array.
    struct page *used_pages[];		//!< Page information for each of the mmaped pages.
} FI_PCI_VMA;

//-----------------------------------------------------------------------------
// Variable Declarations
//-----------------------------------------------------------------------------

//! What BAR do we use for I/O?
extern int fi_bar_select;

//-----------------------------------------------------------------------------
// Function Prototypes
//-----------------------------------------------------------------------------
static long fi_pci_ioctl(struct file *filp, unsigned int cmd, unsigned long arg);
static loff_t   fi_pci_llseek(struct file *filp, loff_t off, int whence);
static int  fi_pci_mmap(struct file *filp, struct vm_area_struct *vma);
static int  fi_pci_open(struct inode *inode, struct file *filp);
static ssize_t  fi_pci_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos);
static int  fi_pci_release(struct inode *inode, struct file *filp);
static int  fi_pci_uevent(struct device *dev, struct kobj_uevent_env *env);
static ssize_t  fi_pci_write(struct file *filp, const char __user *buf, size_t count, loff_t *f_pos);

/*!
 * @brief Used by our class, we're telling the filesystem the permissions
 *    to put on the /dev/fi_pci node.
 */
static int fi_pci_uevent(struct device *dev, struct kobj_uevent_env *env)
{
    add_uevent_var(env, "DEVMODE=%#o", 0666);
    return 0;
}

/*!
 * @brief The ioctl() implementation
 */
static long fi_pci_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
    int retval = -ENOTTY;
    printk(KERN_DEBUG "%s: Not supported!", __FUNCTION__);
    return retval;
}

/*!
 * @brief The llseek() implementation
 */
static loff_t fi_pci_llseek(struct file *filp, loff_t off, int whence)
{
    loff_t newpos;
    FASTEC_PCI_PRIVATE *priv_ptr;       // Pointer to our device structure.

    // Get our pointers.
    priv_ptr = filp->private_data;
    if (priv_ptr == NULL)
    {
        printk(KERN_ERR "%s - No user pointer?!\n", __FUNCTION__);
        return -EFAULT;
    }
    dev_info(&priv_ptr->pci_dev->dev, "%s: whence=%d, off=0x%x\n", __FUNCTION__, whence, (uint32_t)off);

#if 0
    // Lock access to the device.
    if (down_interruptible(&dev->part_sem))
        return -ERESTARTSYS;
#endif

    // Where are we seeking to/from?
    switch (whence)
    {
    case 0: /* SEEK_SET */
        newpos = off;
        break;

    case 1: /* SEEK_CUR */
        newpos = filp->f_pos + off;
        break;

    case 2: /* SEEK_END */
        newpos = priv_ptr->bar_length[fi_bar_select] + off;
        break;

    default: /* can't happen */
        newpos = -EINVAL;
        goto fis_out;
    }

    // Make sure the new position makes sense.
    if ((newpos < 0) || (newpos >= priv_ptr->bar_length[fi_bar_select]))
    {
        printk(KERN_ERR "%s: Invalid new position (0x%x)!\n", __FUNCTION__, (uint32_t)newpos);
        newpos = -EINVAL;
    }
    else
        filp->f_pos = newpos;

fis_out:
#if 0
    // Release the semaphore and return our value.
    up(&dev->part_sem);
#endif
    return newpos;
}

/*!
 * @brief Called to indicate the start of a mmap() operation.
 *
 * For each mmap() call issued from user space, a vm_area_struct is
 * created. This routine is called to indicate that a mmap() operation
 * has started. 
 */
static void fi_pci_vma_open(struct vm_area_struct *vma)
{
    FI_PCI_VMA *pv = vma->vm_private_data;

    atomic_inc(&pv->mmap_count);
}

/*!
 * @brief Release any used pages.
 */
static void fi_pci_vma_free_pages(struct vm_area_struct *vma)
{
    int i;
    FI_PCI_VMA *pv = vma->vm_private_data;

    mutex_lock(&pv->page_mutex);

    for (i = 0; i < pv->nr_pages; i++)
    {
        if (pv->used_pages[i])
        {
            // GENPOOL: fi_pci_free_page(pv->fi_pci_dev, pv->used_pages[i]);
            pv->used_pages[i] = NULL;
        }
    }

    mutex_unlock(&pv->page_mutex);
}

/*!
 * @brief Called by the kernel when an area is destroyed.
 */
static void fi_pci_vma_close(struct vm_area_struct *vma)
{
    FI_PCI_VMA *pv = vma->vm_private_data;

    if (!atomic_dec_and_test(&pv->mmap_count))
        return;

    fi_pci_vma_free_pages(vma);

    dev_dbg(&pv->fi_dev->pci_dev->dev, "vma close");
    kfree(pv);
}

/*!
 * @brief The purpose of this routine is to arrange for a missing
 * 	  page to exist and return its address to the kernel.
 *
 * @param vmf	Pointer to the fault information structure.
 * @returns 	VM_FAULT_OOM if we can't lock the page, or VM_FAULT_NOPAGE
 * 		if the page has been installed in the page table.
 */
static vm_fault_t fi_pci_vma_fault(struct vm_fault *vmf)
{
    FI_PCI_VMA *pv = vmf->vma->vm_private_data;
    unsigned int pg_idx;
    struct page *pg;
    pfn_t pfn;
    int rc;

#if 0
    // Check to see if we're still alive.
    if (!pv->fi_dev->alive)
        return VM_FAULT_SIGBUS;
#endif

    // Figure out what page we're looking for.
    pg_idx = (vmf->address - vmf->vma->vm_start) / PAGE_SIZE;

    // Lock access to our structure.
    mutex_lock(&pv->page_mutex);

    // Have we already locked the page?
    if (pv->used_pages[pg_idx])
        pg = pv->used_pages[pg_idx];
    else
    {
        // dev_info(&pv->fi_dev->pci_dev->dev, "%s: pg_idx=%d\n", __FUNCTION__, pg_idx);
        pg = virt_to_page(&pv->fi_dev->virt_bar_addr[pg_idx * PAGE_SIZE]);
    }

    // Sanity check. Make sure we got a page.
    if (!pg)
    {
        mutex_unlock(&pv->page_mutex);
        return VM_FAULT_OOM;
    }

    // Remember the page information.
    pv->used_pages[pg_idx] = pg;

    // Map the page and add it to our VM list.
    pfn = phys_to_pfn_t(page_to_phys(pg), PFN_DEV | PFN_MAP);
    rc = vmf_insert_mixed(vmf->vma, vmf->address, pfn);

    // Release our lock.
    mutex_unlock(&pv->page_mutex);

    // Return an appropriate status.
    if (rc == -ENOMEM)
        return VM_FAULT_OOM;
    if (rc < 0 && rc != -EBUSY)
        return VM_FAULT_SIGBUS;

    return VM_FAULT_NOPAGE;
}

/*!
 * @brief This describes a s4et of functions that the kernel may
 * 	  invoke to operate on a memory area.
 */
const struct vm_operations_struct fi_pci_vmops =
{
    .open = fi_pci_vma_open,
    .close = fi_pci_vma_close,
    .fault = fi_pci_vma_fault,
};

/*!
 * @brief The mmap() implementation
 */
static int fi_pci_mmap(struct file *filp, struct vm_area_struct *vma)
{
    FI_PCI_VMA *pv;
    FASTEC_PCI_PRIVATE *priv_ptr;       // Pointer to our device structure.
    size_t nr_pages = (vma->vm_end - vma->vm_start) / PAGE_SIZE;

    // Get our pointer.
    priv_ptr = filp->private_data;

    // Make sure we're a shared mapping.
    if ((vma->vm_flags & VM_MAYSHARE) != VM_MAYSHARE)
    {
        dev_warn(&priv_ptr->pci_dev->dev, "mmap failed: can't create private mapping\n");
        return -EINVAL;
    }

    // Allocate our control structure.
    dev_dbg(&priv_ptr->pci_dev->dev, "Allocating mmap with %zd pages.\n", nr_pages);
    pv = kzalloc(sizeof(*pv) + (sizeof(pv->used_pages[0]) * nr_pages), GFP_KERNEL);
    if (!pv)
    {
        dev_err(&priv_ptr->pci_dev->dev, "Unable to allocate memory!\n");
        return -ENOMEM;
    }

    // Initialize our newly allocated structure.
    mutex_init(&pv->page_mutex);
    pv->nr_pages = nr_pages;
    pv->fi_dev = priv_ptr;
    atomic_set(&pv->mmap_count, 1);

    // Hook things in.
    vma->vm_private_data = pv;
    vma->vm_ops = &fi_pci_vmops;
    vma->vm_flags |= VM_MIXEDMAP;

    // All is well.
    return 0;
}

/*!
 * @brief Called when the user opens the device.
 */
static int fi_pci_open(struct inode *inode, struct file *filp)
{
    FASTEC_PCI_PRIVATE *priv_ptr;       // Pointer to our device structure.

    // Get a pointer to our device structure.
    priv_ptr = container_of(inode->i_cdev, FASTEC_PCI_PRIVATE, pci_cdev);
    if (priv_ptr == NULL)
    {
        printk(KERN_ERR "%s - NULL device!\n", __FUNCTION__);
        return -ENODEV;
    }
    filp->private_data = priv_ptr;

    dev_info(&priv_ptr->pci_dev->dev, "%s\n", __FUNCTION__);

    // All done.
    return 0;
}

/*!
 * @brief The read() implementation
 */
static ssize_t fi_pci_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos)
{
    FASTEC_PCI_PRIVATE *priv_ptr;       // Pointer to our device structure.

    // Get our pointer.
    priv_ptr = filp->private_data;
    dev_info(&priv_ptr->pci_dev->dev, "%s: count=%lu\n", __FUNCTION__, count);
    return count;
}

/*!
 * @brief Called when the user closes the device.
 */
static int fi_pci_release(struct inode *inode, struct file *filp)
{
    FASTEC_PCI_PRIVATE *priv_ptr;       // Pointer to our device structure.

    // Get our pointer.
    priv_ptr = filp->private_data;
    dev_info(&priv_ptr->pci_dev->dev, "%s\n", __FUNCTION__);

    // All done.
    return 0;
}

/*!
 * @brief The write() implementation
 */
static ssize_t fi_pci_write(struct file *filp, const char __user *buf, size_t count, loff_t *f_pos)
{
    FASTEC_PCI_PRIVATE *priv_ptr;       // Pointer to our device structure.

    // Get our pointer.
    priv_ptr = filp->private_data;
    dev_info(&priv_ptr->pci_dev->dev, "%s: count=%lu\n", __FUNCTION__, count);
    return count;
}

//! Our file operations structure.
const struct file_operations fi_pci_fops =
{
    .owner =          THIS_MODULE,
    .unlocked_ioctl = fi_pci_ioctl,
    .llseek =         fi_pci_llseek,
    .mmap =           fi_pci_mmap,
    .open =           fi_pci_open,
    .read =           fi_pci_read,
    .release =        fi_pci_release,
    .write =          fi_pci_write,
};

/*!
 * @brief Set up our char device.
 */
int __init fi_pci_cdev_init(FASTEC_PCI_PRIVATE *priv_ptr)
{
    int   retval;

    dev_info(&priv_ptr->pci_dev->dev, "%s\n", __FUNCTION__);

    // Dynamically allocate our major number with only one minor number.
    retval = alloc_chrdev_region(&priv_ptr->devno, 0, MAX_MINOR, MODULE_NAME);
    if (retval < 0)
        return retval;

    // Initialize our cdev structure.
    cdev_init(&priv_ptr->pci_cdev, &fi_pci_fops);

    // Assign some of the fields.
    priv_ptr->pci_cdev.owner = THIS_MODULE;
    priv_ptr->pci_cdev.ops = &fi_pci_fops;

    // Create a unique class because there's no one like us.
    priv_ptr->pci_class = class_create(THIS_MODULE, MODULE_NAME);
    if (IS_ERR(priv_ptr->pci_class))
    {
        retval = PTR_ERR(priv_ptr->pci_class);
        goto err1;
    }

    // Hook in our uevent handler so the /dev/fi_pci node will have
    // the correct permissions.
    priv_ptr->pci_class->dev_uevent = fi_pci_uevent;

    // Create the device for this class.
    priv_ptr->pci_cdev_dev = device_create(priv_ptr->pci_class, NULL, priv_ptr->devno, NULL, MODULE_NAME);
    if (IS_ERR(priv_ptr->pci_cdev_dev))
    {
        retval = PTR_ERR(priv_ptr->pci_cdev_dev);
        goto err2;
    }

    // Add this device to the list.
    retval = cdev_add(&priv_ptr->pci_cdev, priv_ptr->devno, MAX_MINOR);
    if (retval < 0)
    {
        cdev_del(&priv_ptr->pci_cdev);
        goto err3;
    }

    return 0;

err3:
    device_destroy(priv_ptr->pci_class, priv_ptr->devno);
    priv_ptr->pci_cdev_dev = NULL;
err2:
    class_destroy(priv_ptr->pci_class);
    priv_ptr->pci_class = NULL;
err1:
    unregister_chrdev_region(priv_ptr->devno, MAX_MINOR);

    // Return our status.
    return retval;
}

/*!
 * @brief Clean up the character device on exit.
 */
void __exit fi_pci_cdev_exit(FASTEC_PCI_PRIVATE *priv_ptr)
{
    // Unregister the userland interface.
    cdev_del(&priv_ptr->pci_cdev);
    device_destroy(priv_ptr->pci_class, priv_ptr->devno);
    class_destroy(priv_ptr->pci_class);
    unregister_chrdev_region(priv_ptr->devno, MAX_MINOR);
}

